import unittest
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression, LinearRegression

from causallib.metrics import weighted_roc_auc_error, expected_roc_auc_error
from causallib.metrics import weighted_roc_curve_error, expected_roc_curve_error
from causallib.metrics import ici_error
from causallib.metrics import covariate_balancing_error
from causallib.metrics import balanced_residuals_error


class TestPropensityMetrics(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Data:
        X, a = make_classification(
            n_features=1, n_informative=1, n_redundant=0, n_repeated=0,
            n_classes=2, n_clusters_per_class=1, flip_y=0.0, class_sep=0.5,
            random_state=0,
        )
        cls.data_r_100 = {"X": pd.DataFrame(X), "a": pd.Series(a)}

        # RCT Data:
        np.random.seed(0)
        X = np.random.normal(0, 1, 10000)
        a = np.random.binomial(1, 0.5, 10000)
        cls.data_rct = {"X": pd.DataFrame(X), "a": pd.Series(a)}

        # Hard-coded Data:
        y = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        y_pred = pd.Series([0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 1, 1])
        y_chance = pd.Series(10 * [0.5])
        w_uniform = pd.Series(10 * [0.1])
        w = pd.Series([1 / 6, 1 / 6, 1 / 6, 1 / 6, 1 / 6, 1 / 6, 0, 0, 0, 0])

        cls.hard_coded_data = {
            "y": y,
            "y_chance": y_chance,
            "y_pred": y_pred,
            "w_uniform": w_uniform,
            "w": w
        }

        # # Avoids regularization of the model:
        cls.estimator = LogisticRegression(penalty='none', solver='sag', max_iter=2000)

    def test_weighted_roc_auc(self):
        with self.subTest("Chance predictions"):
            expected = 0
            observed = weighted_roc_auc_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_chance'],
                sample_weight=self.hard_coded_data['w_uniform']
            )
            self.assertEqual(observed, expected)

        with self.subTest("Random label fitting"):
            self.estimator.fit(self.data_rct['X'], self.data_rct['a'])
            a_pred = self.estimator.predict_proba(self.data_rct['X'])
            ip_weights = 1 / a_pred[range(a_pred.shape[0]), self.data_rct['a']]
            expected = 0
            observed = weighted_roc_auc_error(
                y_true=self.data_rct['a'],
                y_pred=a_pred[:, 1],
                sample_weight=ip_weights
            )
            self.assertAlmostEqual(observed, expected, places=6)

        with self.subTest("Good separation"):
            self.estimator.fit(self.data_r_100['X'], self.data_r_100['a'])
            a_pred = self.estimator.predict_proba(self.data_r_100['X'])
            ip_weights = 1 / a_pred[range(a_pred.shape[0]), self.data_r_100['a']]
            expected = 0.25  # The result for this data
            observed = weighted_roc_auc_error(
                y_true=self.data_r_100['a'],
                y_pred=a_pred[:, 1],
                sample_weight=ip_weights
            )
            self.assertAlmostEqual(observed, expected, places=4)

        with self.subTest("Perfect prediction but uniform weights"):
            # When the weights are uniform, i.e. weight_sample=1/n, W-AUC is
            # similar to AUC. Here, the prediction is perfect and the
            # W-AUC=AUC=1. The squared error is (W-AUC-0.5)**2=0.25
            expected = 0.25
            observed = weighted_roc_auc_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y'],
                sample_weight=self.hard_coded_data['w_uniform']
            )
            self.assertEqual(observed, expected)

        with self.subTest("Uniform weighted AUC"):
            # W-AUC and AUC are similar, but smaller than 1.
            weighted_auc_uniform = weighted_roc_auc_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_pred'],
                sample_weight=self.hard_coded_data['w_uniform']
            )
            self.assertLess(weighted_auc_uniform, 0.25)

        with self.subTest("Non-uniform weights should be closer to the diagonal"):
            # The weights are not uniform, and the W-AUC is closer to the diagonal
            weighted_auc = weighted_roc_auc_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_pred'],
                sample_weight=self.hard_coded_data['w']
            )
            self.assertLess(weighted_auc, weighted_auc_uniform)

    def test_expected_roc_auc_error(self):
        with self.subTest("Identical prediction, perfect alignment"):
            expected_roc_auc = expected_roc_auc_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y']
            )
            self.assertEqual(expected_roc_auc, 0.0)

        with self.subTest("Good calibration"):
            expected_roc_auc = expected_roc_auc_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_pred']
            )
            self.assertAlmostEqual(0, expected_roc_auc, places=2)

    def test_weighted_roc_curve(self):
        with self.subTest("Chance predictions"):
            expected = 0
            observed = weighted_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_chance'],
                sample_weight=self.hard_coded_data['w_uniform']
            )
            self.assertEqual(observed, expected)

        with self.subTest("Random label fitting"):
            self.estimator.fit(self.data_rct['X'], self.data_rct['a'])
            a_pred = self.estimator.predict_proba(self.data_rct['X'])
            ip_weights = 1 / a_pred[range(a_pred.shape[0]), self.data_rct['a']]
            expected = 0  # 0.0130
            observed = weighted_roc_curve_error(
                y_true=self.data_rct['a'],
                y_pred=a_pred[:, 1],
                sample_weight=ip_weights
            )
            self.assertAlmostEqual(observed, expected, places=1)

        with self.subTest("Perfect separation"):
            self.estimator.fit(self.data_r_100['X'], self.data_r_100['a'])
            a_pred = self.estimator.predict_proba(self.data_r_100['X'])
            ip_weights = 1 / a_pred[range(a_pred.shape[0]), self.data_r_100['a']]
            expected = 1.0
            observed = weighted_roc_curve_error(
                y_true=self.data_r_100['a'],
                y_pred=a_pred[:, 1],
                sample_weight=ip_weights
            )
            self.assertAlmostEqual(observed, expected, places=4)

        with self.subTest("Perfect prediction but uniform weights"):
            # perfect prediction so the difference from diagonal is 1 (top-left minus bottom-left corner)
            expected = 1.0
            observed = weighted_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y'],
                sample_weight=self.hard_coded_data['w_uniform']
            )
            self.assertEqual(observed, expected)

        with self.subTest("Uniform weighted ROC curve"):
            expected = 0.4  # The result from this `y_pred` assignment
            weighted_auc_uniform = weighted_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_pred'],
                sample_weight=self.hard_coded_data['w_uniform']
            )
            self.assertEqual(weighted_auc_uniform, expected)

        with self.subTest("Non-uniform weights should be closer to the diagonal"):
            # The weights are not uniform, and the W-AUC is closer to the diagonal
            weighted_auc = weighted_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_pred'],
                sample_weight=self.hard_coded_data['w']
            )
            self.assertLess(weighted_auc, weighted_auc_uniform)

    def test_expected_roc_curve_error(self):
        with self.subTest("Identical prediction, perfect alignment"):
            expected_roc_auc = expected_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y']
            )
            self.assertEqual(expected_roc_auc, 0.0)

        with self.subTest("Good calibration"):
            expected_roc_auc = expected_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y_pred']
            )
            expected = 0.0539
            self.assertAlmostEqual(expected_roc_auc, expected, places=3)

    def test_different_agg_function_in_curve(self):
        with self.subTest("Weighted ROC curve"):
            # perfect prediction so the difference from diagonal is 1 (top-left minus bottom-left corner)
            expected = 1/3  # 3-point curve: bottom-left (0 difference), top-left (1), top-right (0)
            observed = weighted_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y'],
                sample_weight=self.hard_coded_data['w_uniform'],
                agg=np.mean
            )
            self.assertEqual(observed, expected)

        with self.subTest("Expected ROC curve"):
            expected = 0.0  # Equal curve so diff=0 all the way
            observed = expected_roc_curve_error(
                y_true=self.hard_coded_data['y'],
                y_pred=self.hard_coded_data['y'],
                agg=np.mean,
            )
            self.assertEqual(observed, expected)

    def test_ici(self):
        data = {  # Reduce size since lowess can be slow
            'X': self.data_rct['X'].loc[:200],
            'a': self.data_rct['a'].loc[:200],
        }
        self.estimator.fit(data['X'], data['a'])
        a_pred = self.estimator.predict_proba(data['X'])[:, 1]

        with self.subTest("Default parameters"):
            ici_score = ici_error(data['a'], a_pred)
            expected = 0.0239
            self.assertAlmostEqual(ici_score, expected, places=3)

        with self.subTest("`return_sorted==False`"):
            ici_score = ici_error(
                data['a'], a_pred,
                lowess_kwargs={"return_sorted": False}
            )
            expected = 0.0239
            self.assertAlmostEqual(ici_score, expected, places=3)


class TestWeightMetrics(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        a = pd.Series([0]*6 + [1]*6)
        X = pd.DataFrame(
            {
                "x1": [1, 0] * 6,  # Perfectly balanced
                "x2": [0]*5 + [1, 0] + [1]*5,  # Almost perfectly imbalance (pure imbalance crashes when standardizing)
            }
        )
        w = pd.Series(data=1/6, index=a.index)
        cls.data = {"X": X, "a": a, "w": w}

        # # Avoids regularization of the model:
        cls.estimator = LogisticRegression(penalty='none', solver='sag', max_iter=2000)

    def test_covariate_balancing(self):
        score = covariate_balancing_error(self.data["X"], self.data["a"], self.data["w"])
        expected = (5/6 - 1/6) / np.sqrt(0.13888 + 0.13888)
        self.assertAlmostEqual(score, expected, places=4)

        with self.subTest("Test `mean` agg function"):
            score = covariate_balancing_error(
                self.data["X"], self.data["a"], self.data["w"],
                agg=np.mean,
            )
            expected /= 2  # Two features, the second has 0 ASMD
            self.assertAlmostEqual(score, expected, places=4)


class TestOutcomeMetrics(unittest.TestCase):
    def test_balanced_residuals(self):
        a = pd.Series([0]*5 + [1]*6)  # Unequal group sizes
        y = pd.Series([8, 9, 10, 11, 12,
                       18, 19, 20, 20, 21, 22])
        y_pred = pd.DataFrame({  # Will only use the observed predictions
            0: [10]*5 + [np.nan]*6,
            1: [np.nan]*5 + [20]*6,
        })

        score = balanced_residuals_error(y, y_pred, a)
        expected = 0
        self.assertAlmostEqual(score, expected)

        with self.subTest("Custom distance metric"):
            from scipy.stats import ks_2samp
            score = balanced_residuals_error(
                y, y_pred, a,
                distance_metric=lambda y0, y1: ks_2samp(y1, y0)[0]  # Return KS statistic
            )
            expected = 0.06666666666  # Result per KS
            self.assertAlmostEqual(score, expected)

        with self.subTest("Biased residuals"):
            bias = 5
            y_pred_bias = y_pred.copy()
            y_pred_bias.loc[a == 1] += bias
            # Simple un-standardized mean-difference should reproduce the bias
            score = balanced_residuals_error(
                y, y_pred_bias, a,
                distance_metric=lambda y1, y0: abs(y1.mean() - y0.mean())
            )
            expected = bias
            self.assertEqual(score, expected)


# if __name__ == '__main__':
#     unittest.main()



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Feb 21, 2019

"""

import unittest

from causallib.utils import general_tools
from causallib.utils.stat_utils import robust_lookup


class TestUtils(unittest.TestCase):
    def ensure_learner_is_fitted(self, unfitted_learner, X, y):
        model_name = str(unfitted_learner.__class__).split(".")[-1]
        with self.subTest("Check is_fitted of {}".format(model_name)):
            self.assertFalse(general_tools.check_learner_is_fitted(unfitted_learner))
            unfitted_learner.fit(X, y)
            self.assertTrue(general_tools.check_learner_is_fitted(unfitted_learner))

    def test_check_classification_learner_is_fitted(self):
        from sklearn.linear_model import LogisticRegression
        from sklearn.tree import DecisionTreeClassifier
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.svm import LinearSVC
        from sklearn.datasets import make_classification
        X, y = make_classification()
        for clf in [LogisticRegression(solver='lbfgs'), DecisionTreeClassifier(),
                    RandomForestClassifier(), LinearSVC()]:
            self.ensure_learner_is_fitted(clf, X, y)

    def test_check_regression_learner_is_fitted(self):
        from sklearn.linear_model import LinearRegression
        from sklearn.tree import ExtraTreeRegressor
        from sklearn.ensemble import GradientBoostingRegressor
        from sklearn.svm import SVR
        from sklearn.datasets import make_regression
        X, y = make_regression()
        for regr in [LinearRegression(), ExtraTreeRegressor(),
                     GradientBoostingRegressor(), SVR()]:
            self.ensure_learner_is_fitted(regr, X, y)

    def test_robust_lookup(self):
        import pandas as pd
        with self.subTest("zero-one columns, range index"):
            X = pd.DataFrame({
                0: [10, 20, 30, 40],
                1: [100, 200, 300, 400]
            })
            a = pd.Series([0, 1, 0, 1])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("integer columns, range index"):
            X = pd.DataFrame({
                3: [10, 20, 30, 40],
                4: [100, 200, 300, 400]
            })
            a = pd.Series([3, 4, 3, 4])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("text columns, range index"):
            X = pd.DataFrame({
                "a": [10, 20, 30, 40],
                "b": [100, 200, 300, 400]
            })
            a = pd.Series(["a", "b", "a", "b"])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("text columns, text index"):
            X = pd.DataFrame({
                "a": [10, 20, 30, 40],
                "b": [100, 200, 300, 400]
            }, index=["w", "x", "y", "z"])
            a = pd.Series(["a", "b", "a", "b"], index=["w", "x", "y", "z"])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400], index=["w", "x", "y", "z"])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("index mismatch"):
            X = pd.DataFrame({
                0: [10, 20, 30, 40],
                1: [100, 200, 300, 400]
            }, index=[0, 1, 2, 3])
            a = pd.Series([0, 1, 0, 1], index=[0, 1, 2, 4])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, None], index=[0, 1, 2, 4])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("columns mismatch"):
            X = pd.DataFrame({
                0: [10, 20, 30, 40],
                1: [100, 200, 300, 400]
            })
            a = pd.Series([0, 1, 0, 2])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, None])
            pd.testing.assert_series_equal(expected, extracted)



import pandas as pd
import numpy as np
from typing import List, Optional, Union
import uuid
from collections import Counter


def get_person_time_df(t: pd.Series, y: pd.Series, a: Optional[pd.Series] = None, w: Optional[pd.Series] = None,
                       X: Optional[pd.DataFrame] = None, return_individual_series: bool = False) -> pd.DataFrame:
    """
    Converts standard input format into an expanded person-time format.
    Input series need to be indexed by subject IDs and have non-null names (including index name).

    Args:
        t (pd.Series): Followup duration, size (num_subjects,).
        y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
        a (pd.Series): Treatment assignment of size (num_subjects,).
        w (pd.Series): Optional subject weights
        X (pd.DataFrame): Optional baseline covariate matrix of size (num_subjects, num_features).
        return_individual_series (bool): If True, returns a tuple of Series/DataFrames instead of a single DataFrame

    Returns:
        pd.DataFrame: Expanded person-time format with columns from X and expanded 'a', 'y', 't' columns

    Examples:
        This example standard input:

            age  height  a  y  t
        id
        1    22     170  0  1  2
        2    40     180  1  0  1
        3    30     165  1  0  2


        Will be expanded to:

            age  height  a  y  t
        id
        1    22     170  0  0  0
        1    22     170  0  0  1
        1    22     170  0  1  2
        2    40     180  1  0  0
        2    40     180  1  0  1
        3    30     165  1  0  0
        3    30     165  1  0  1
        3    30     165  1  0  2

    """
    a, t, y, w, X = canonize_dtypes_and_names(a, t, y, w, X)
    data = pd.concat([X, w, a], axis=1)  # X and/or w may be None - in which case they're ignored by pandas

    result_col_order = list(data.columns) + [y.name, t.name]

    # Create discrete-increment time vector according to the time of each id (index):
    t_expand = t.reindex(t.index.repeat(t + 1))  # Replicate rows by subject time (+1 since it's a closed-interval)
    t_expand = t_expand.groupby(t_expand.index).cumcount()  # filling each row with time from 0 to t_id
    t_expand.name = t.name

    # prepare `y` to merge on `id` and time
    yt = pd.concat([y, t], axis=1)

    # merge outcome with unfolded-time, so that the outcome only matches the existing (last) event time,
    yt_expand = pd.merge(left=t_expand.reset_index(), right=yt.reset_index(), how="left", on=[t.index.name, t.name])

    # wherever `y` didn't match for the `id, t` combo, it was set to NaN - so impute with zero
    yt_expand[y.name] = yt_expand[y.name].fillna(0).astype(int)

    # merge covariates-treatment on `id` will simply duplicate the covariates/treatment to fit the expanded format
    res = pd.merge(left=data, right=yt_expand, how='right', left_index=True, right_on=data.index.name)
    res.set_index(t.index.name, inplace=True)  # restore index by subject id

    if return_individual_series:
        # split into X, w, a, y and t (all in person-time format)
        res = (
            res[X.columns] if X is not None else None,
            res[w.name] if w is not None else None,
            res[a.name] if a is not None else None,
            res[t.name],
            res[y.name],
        )
    else:
        res = res[result_col_order]  # canonize columns order: covariate columns, w, a, y, t
    return res


def get_regression_predict_data(X: pd.DataFrame, times: pd.Series):
    """
    Generates prediction data for a regression fitter: repeats patient covariates per time point in 'times'.
    Example:

        X
        ===================
                age  height
            id
            1    22     170
            2    40     180

        times
        =====================
            0
            1
            2

        Result - pred_data_X
        =====================
                age  height  t
            id
            1    22     170  0
            1    22     170  1
            1    22     170  2
            2    40     180  0
            2    40     180  1
            2    40     180  2

    Args:
        X (pd.DataFrame):  Covariates DataFrame
        times (pd.Series):  A Series of time points to predict

    Returns:
        pred_data_X (pd.DataFrame):  DataFrame with repeated covariates per time point.
                                     Index is subject ID with repeats, columns are X + a time column, which is a repeat
                                     of 'times' per subject.
        t_name (str):  Name of time column in pred_data_X. Default is 't', but since we concatenate a column to a
                       covariates frame, we might need to add a random suffix to it.

    """
    X_repeated = X.loc[X.index.repeat(len(times))]
    times_repeated = pd.concat([times] * X.shape[0], axis=0)
    times_repeated.index = X_repeated.index
    times_repeated.name = times.name if times.name is not None else 't'
    times_repeated = times_repeated.astype('int64')

    # Pred_data_X is a cartesian product of times and features (X rows are repeated per time point)
    pred_data_X, new_col_names = safe_join(df=X_repeated, list_of_series=[times_repeated], return_series_names=True)
    t_name = new_col_names[0]

    return pred_data_X, t_name


def compute_survival_from_single_hazard_curve(hazard: List, logspace: bool = False) -> List:
    """
    Computes survival curve from an array of point hazards.
    Note that trailing NaN are supported
    Args:
        hazard (list):  list/array of point hazards
        logspace (bool):  whether to compute in logspace, for numerical stability

    Returns:
        list: survival at each time-step
    """
    survival = 1 - hazard
    if logspace:
        log_survival = np.log(survival)
        log_survival = np.cumsum(log_survival)
        survival = np.exp(log_survival)
    else:
        survival = np.cumprod(survival)

    return survival


def canonize_dtypes_and_names(a=None, t=None, y=None, w=None, X=None):
    """
    Housekeeping method that assign names for unnamed series and canonizes their data types.

    Args:
        a (pd.Series|None): Treatment assignment of size (num_subjects,).
        t (pd.Series|None): Followup duration, size (num_subjects,).
        y (pd.Series|None): Observed outcome (1) or right censoring event (0), size (num_subjects,).
        w (pd.Series|None): Optional subject weights
        X (pd.DataFrame|None): Baseline covariate matrix of size (num_subjects, num_features).

    Returns:
        a, y, t, w, X
    """
    if a is not None:
        a = a.astype(int)
        a.name = 'a' if a.name is None else a.name

    if t is not None:
        t = t.astype(int)
        t.name = 't' if t.name is None else t.name

    if y is not None:
        y = y.astype(int)
        y.name = 'y' if y.name is None else y.name

    if w is not None:
        w.name = 'w' if w.name is None else w.name

    # Set index name
    for obj in [a, y, t, w, X]:
        if obj is not None and obj.index.name is None:
            obj.index.name = 'id'

    return a, t, y, w, X


def add_random_suffix(name, suffix_length=4):
    """
    Adds a random suffix to string, by computing uuid64.hex.

    Args:
        name: input string
        suffix_length: length of desired added suffix.

    Returns:
        string with suffix
    """
    if suffix_length:
        name += "_" + uuid.uuid4().hex[:suffix_length]
    return name


def safe_join(df: Optional[pd.DataFrame] = None, list_of_series: List[pd.Series] = None, return_series_names=False):
    """
    Safely joins (concatenates on axis 1) a collection of Series (or one DataFrame and multiple Series),
    while renaming Series that have a duplicate name (a name that already exists in DataFrame or another Series).
    * Note that DataFrame columns are never changed (only Series names are).

    Args:
        df (pd.DataFrame):  optional DataFrame. If provided, will join Series to DataFrame
        list_of_series (List[pd.Series]): list of Series for safe-join
        return_series_names (bool):  if True, returns a list of (potentially renamed) Series names

    Returns:
        1. single concatenated DataFrame
        2. list of (potentially renamed) Series names
    """

    list_of_frames_or_series: List[Union[pd.DataFrame, pd.Series]] = []
    if df is not None:
        list_of_frames_or_series.append(df)
    list_of_frames_or_series.extend(list_of_series)

    # Get all column names
    all_col_names = []
    for item in list_of_frames_or_series:
        if isinstance(item, pd.Series):
            all_col_names.append(item.name)
        elif isinstance(item, pd.DataFrame):
            all_col_names.extend(item.columns)

    duplicate_col_names = [col_name for col_name, count in Counter(all_col_names).items() if count > 1]

    # If no duplicates, concat without renaming
    if len(duplicate_col_names) == 0:
        concat_res = pd.concat(list_of_frames_or_series, axis=1)
        new_names = [s.name for s in list_of_series]

    # Rename Series with duplicate name
    else:
        list_of_renamed_series = []
        new_names = []
        for series in list_of_series:
            if series.name in duplicate_col_names or series.name is None:
                orig_name = str(series.name)
                new_name = add_random_suffix(orig_name)
                while new_name in all_col_names:
                    new_name = add_random_suffix(orig_name)
                series_copy = series.copy()
                series_copy.name = new_name

                list_of_renamed_series.append(series_copy)
                new_names.append(new_name)
            else:
                list_of_renamed_series.append(series)
                new_names.append(series.name)

        concat_res = pd.concat([df] + list_of_renamed_series, axis=1)

    if return_series_names:
        return concat_res, new_names
    else:
        return concat_res



# Implements Shared Sparsity Confounder Selection suggested by:
# https://arxiv.org/abs/2011.01979

import warnings
import numpy as np
from sklearn.exceptions import ConvergenceWarning
from causallib.preprocessing.confounder_selection import _BaseConfounderSelection

__all__ = ["SharedSparsityConfounderSelection"]


class MCPSelector:
    # TODO: transpose `theta` and rename it `coef_` to align with sklearn models
    def __init__(self, lmda="auto", alpha=1, step=0.1, max_iter=1000, tol=1e-3):
        """Constructor for MCPSelector. This class computes shared
        sparsity matrix using proximal gradient descent applied with
        MCP regularizer.

        Args:
            lmda (str|float): Parameter (>= 0) to control shape of MCP regularizer.
                The bigger the value the stronger the regularization.
                "auto" will auto-select good regularization value.
            alpha (float): Associated lambda parameter (>= 0) to control shape of MCP regularizer.
                The smaller the value the stronger the regularization.
            step (float): Step size for proximal gradient, equivalent of learning rate.
            max_iter (int): Maximum number of iterations of MCP proximal gradient.
            tol (float): Stopping criterion for MCP. If the normalized value of
                proximal gradient is less than tol then the algorithm is assumed
                to have converged.
        """
        super().__init__()
        self.alpha = alpha
        self.lmda = lmda
        self.step = step
        self.max_iter = max_iter
        self.tol = tol
        self.epsilon_safe_division = 1e-6

    def _initialize_internal_params(self, X, a, y):
        treatments = list(a.unique())
        n_treatments = len(treatments)
        n_confounders = X.shape[1]
        if self.lmda == "auto":
            avg_num_samples = len(y) / n_treatments
            lmda = 0.2 * np.sqrt(n_treatments * np.log(n_confounders) / avg_num_samples)
        else:
            lmda = self.lmda

        assert self.alpha >= 0
        assert lmda >= 0

        corr_xx = np.zeros((n_confounders, n_confounders, n_treatments))
        corr_xa = np.zeros((n_confounders, n_treatments))
        for i, t in enumerate(treatments):
            u, v = X.loc[a == t].values.T, y.loc[a == t].values
            corr_xx[:, :, i] = (u @ u.T) / len(v)
            corr_xa[:, i] = (u @ v) / len(v)
        self.theta_ = np.zeros((n_confounders, n_treatments))
        self.corr_mats_ = {"corr_xx": corr_xx, "corr_xa": corr_xa}  # Correlation matrices
        self.n_confounders_ = n_confounders
        self.n_treatments_ = n_treatments
        self.lmda_ = lmda

    def _mcp_q_grad(self, t):
        if self.alpha == 0:
            return np.sign(t) * self.lmda_ * (-1)
        else:
            return (np.sign(t) * self.lmda_
                    * np.maximum(-1, - np.abs(t) / (self.lmda_ * self.alpha)))

    def _update_grad_op(self, theta):
        theta_new = np.zeros_like(theta)
        for j in range(self.n_treatments_):
            sub = (self.corr_mats_["corr_xx"][:, :, j] @ theta[:, j]
                   - self.corr_mats_["corr_xa"][:, j])
            theta_new[:, j] = theta[:, j] - self.step * sub
        norm_matrix = (np.linalg.norm(theta, axis=1, keepdims=True)
                       @ np.ones((1, self.n_treatments_)))
        eps = self.epsilon_safe_division
        theta_grad_op = theta_new - (self.step * (self._mcp_q_grad(norm_matrix))
                                     * theta / (norm_matrix + eps))
        return theta_grad_op

    def _update_proximal_op(self, theta):
        norm_theta = np.linalg.norm(theta, axis=1, keepdims=True)
        shrink = np.maximum(0, norm_theta - self.lmda_)
        theta_proximal_op = theta * ((shrink / norm_theta)
                                     @ np.ones((1, self.n_treatments_)))
        return theta_proximal_op

    def compute_shared_sparsity_matrix(self, X, a, y):
        self._initialize_internal_params(X, a, y)
        theta = self.theta_
        for i in range(self.max_iter):
            theta_prev = theta
            theta_grad_op = self._update_grad_op(theta_prev)
            theta = self._update_proximal_op(theta_grad_op)

            convergence_criteria = (
                    np.linalg.norm(theta - theta_prev) /
                    (np.linalg.norm(theta) + self.epsilon_safe_division)
            )
            has_converged = convergence_criteria <= self.tol
            if has_converged:
                break
        else:
            warnings.warn("Shared sparsity did not converge in maximum iterations.", ConvergenceWarning)
        self.theta_ = theta
        return self.theta_


class SharedSparsityConfounderSelection(_BaseConfounderSelection):
    """Class to select confounders by first applying shared sparsity method.
    Method by Greenewald, Katz-Rogozhnikov, and Shanmugam: https://arxiv.org/abs/2011.01979
    """

    def __init__(self, mcp_lambda="auto", mcp_alpha=1, step=0.1, max_iter=1000, tol=1e-3, threshold=1e-6,
                 importance_getter=None, covariates=None):
        """Constructor for SharedSparsityConfounderSelection

        Args:
            mcp_lambda (str|float): Parameter (>= 0) to control shape of MCP regularizer.
                The bigger the value the stronger the regularization.
                "auto" will auto-select good regularization value.
            mcp_alpha (float): Associated lambda parameter (>= 0) to control shape of MCP regularizer.
                The smaller the value the stronger the regularization.
            step (float): Step size for proximal gradient, equivalent of learning rate.
            max_iter (int): Maximum number of iterations of MCP proximal gradient.
            tol (float): Stopping criterion for MCP. If the normalized value of
                proximal gradient is less than tol then the algorithm is assumed
                to have converged.
            threshold (float): Only if the importance of a confounder exceeds
                threshold for all values of treatments, then the confounder
                is retained by transform() call.
            importance_getter: IGNORED.
            covariates (list | np.ndarray): Specifying a subset of columns to perform selection on.
                Columns in `X` but not in `covariates` will be included after `transform`
                no matter the selection.
                Can be either a list of column names, or an array of boolean indicators length of `X`,
                or anything compatible with pandas `loc` function for columns.
                if `None` then all columns are participating in the selection process.
                This is similar to using sklearn's `ColumnTransformer` or `make_column_selector`.
        """
        super().__init__(importance_getter=importance_getter, covariates=covariates)
        self.step = step
        self.max_iter = max_iter
        self.tol = tol
        self.threshold = threshold
        self.selector_ = MCPSelector(lmda=mcp_lambda, alpha=mcp_alpha, step=step, max_iter=max_iter, tol=tol)

        self.importance_getter = lambda e: e.selector_.theta_.transpose()  # Shape to behave like sklearn linear_model
        # self.importance_getter = "selector_.theta_"

    @_BaseConfounderSelection._filter_covariates
    def fit(self, X, a, y):
        # compute_shared_sparsity_matrix() below should return
        # a matrix of shape (n_confounders x n_treatments).
        # The confounders to be retained in transform() corresponds to those rows
        # which have significant values across all the columns.
        theta = self.selector_.compute_shared_sparsity_matrix(X, a, y)
        support = (np.abs(theta) >= self.threshold).all(axis=1)
        if not support.any():
            warnings.warn("Shared sparsity selected zero features. Ignoring and selecting all features.")
            self.support_ = np.ones(X.shape[1], dtype=bool)
        else:
            self.support_ = support
        self.n_features_ = sum(self.support_)
        return self

    # @_BaseConfounderSelection._filter_and_re_add_covariates
    # def transform(self, X, a=None):
    #     X = X.loc[:, self.get_support()]
    #     return X

    def _get_support_mask(self):
        return self.support_



# coding: utf-8

# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Sept 25, 2019

import torch.nn.functional as F
import torch.nn as nn
import torch


class BalancedNet(nn.Module):
    """A torch.model used as a component of the HEMM module to determine the outcome as a function of confounders. 
    The balanced net consists of two different neural networks for the outcome and counteractual.
    """

    def __init__(self, D_in, H, D_out):
        """Instantiate two nn.Linear modules and assign them as member variables.

        Args:
            D_in: input dimension
            H: dimension of hidden layer
            D_out: output dimension
        """

        super(BalancedNet, self).__init__()
        
        self.f1 = nn.Linear(D_in, H) 
        self.f2 = nn.Linear(H, D_out)
        
        self.cf1 = nn.Linear(D_in, H) 
        self.cf2 = nn.Linear(H, D_out)
  
    def forward(self, x):
        """Accept a Variable of input data and return a Variable of output data.

        We can use Modules defined in the constructor as well as arbitrary operators on Variables.
        """
        h_relu = F.elu(self.f1(x))
        f = self.f2(h_relu)
        
        h_relu = F.elu(self.cf1(x))
        cf = self.cf2(h_relu)
        
        out = torch.cat((f, cf), dim=1)
        
        return out


def genMLPModule(D_in, H, out=1):
    """Fit an MLP with an ELU activation.

    This allows for a single neural network to have two heads for the outcome and counterfactual.
    """
    if type(H) is int:
        model = torch.nn.Sequential(torch.nn.Linear(D_in, H), torch.nn.ELU(), torch.nn.Linear(H, out))
        return model.double()


def genLinearModule(D_in,  out=1):
    """Two separate linear functions of the input covariates."""
    model = torch.nn.Sequential(torch.nn.Linear(D_in, out),)   
    return model.double()



import unittest
import pandas as pd

from causallib.model_selection import TreatmentOutcomeStratifiedKFold
from causallib.model_selection import TreatmentStratifiedKFold


class TestTreatmentOutcomeStratifiedKFold(unittest.TestCase):
    def test_binary_splits(self):
        X = pd.DataFrame({0: [10] * 8})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1])
        y = pd.Series([0, 0, 1, 1, 0, 0, 1, 1])
        kfold = TreatmentOutcomeStratifiedKFold(n_splits=2)
        folds = list(kfold._split(X, a, y))[0]  # 2 folds, so first element suffices

        self.assertListEqual(
            list(folds[0]), [1, 3, 5, 7],
        )
        self.assertListEqual(
            list(folds[1]), [0, 2, 4, 6],
        )

        labels = kfold._combine_treatment_outcome_labels(a, y)
        self.assertListEqual(
            list(labels),
            [0, 0, 1, 1, 2, 2, 3, 3],
        )

    def ensure_multiclass_splits(self, X, a, y):
        kfold = TreatmentOutcomeStratifiedKFold(n_splits=2)
        folds = list(kfold._split(X, a, y))[0]  # 2 folds, so first element suffices

        self.assertListEqual(
            list(folds[0]),  # Train indices
            [1, 3, 5, 7, 9, 11]
        )
        self.assertListEqual(
            list(folds[1]),  # Train indices
            [0, 2, 4, 6, 8, 10]
        )

    def test_multi_outcome_splits(self):
        X = pd.DataFrame({0: [10] * 12})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
        y = pd.Series([0, 0, 1, 1, 2, 2, 0, 0, 1, 1, 2, 2])
        self.ensure_multiclass_splits(X, a, y)

        labels = TreatmentOutcomeStratifiedKFold._combine_treatment_outcome_labels(a, y)
        self.assertListEqual(
            list(labels),
            [0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
        )

    def test_multi_treatment_splits(self):
        X = pd.DataFrame({0: [10] * 12})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])
        y = pd.Series([0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1])
        self.ensure_multiclass_splits(X, a, y)

        labels = TreatmentOutcomeStratifiedKFold._combine_treatment_outcome_labels(a, y)
        self.assertListEqual(
            list(labels),
            [0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
        )

    def test_continuous_outcome(self):
        X = pd.DataFrame({0: [10] * 4})  # Dummy corresponding X
        a = pd.Series([0, 0, 1, 1])
        y = pd.Series([42.0, 42.1, 42.2, 42.3])
        kfold = TreatmentOutcomeStratifiedKFold(n_splits=2, shuffle=False)
        with self.assertRaises(ValueError):
            kfold._split(X, a, y)

    def test_sklearn_search_compatibility(self):
        X = pd.DataFrame({0: [10] * 8})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1])
        y = pd.Series([0, 0, 1, 1, 0, 0, 1, 1])
        Xa = X.join(a.to_frame("a"))
        kfold = TreatmentOutcomeStratifiedKFold(n_splits=2)
        folds = list(kfold.split(Xa, y))[0]  # 2 folds, so first element suffices

        self.assertListEqual(
            list(folds[0]), [1, 3, 5, 7],
        )
        self.assertListEqual(
            list(folds[1]), [0, 2, 4, 6],
        )


class TestTreatmentStratifiedKFold(unittest.TestCase):
    def test_binary_splits(self):
        X = pd.DataFrame({0: [10] * 8})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1])
        y = pd.Series([0, 0, 1, 1, 0, 0, 1, 1])  # should be ignored
        kfold = TreatmentStratifiedKFold(n_splits=2)
        folds = list(kfold._split(X, a, y))[1]  # 2 folds, so first element suffices

        self.assertListEqual(
            list(folds[0]), [0, 1, 4, 5],
        )
        self.assertListEqual(
            list(folds[1]), [2, 3, 6, 7],
        )

    def test_with_no_outcome(self):
        X = pd.DataFrame({0: [10] * 8})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1])
        Xa = X.join(a.to_frame("a"))
        kfold = TreatmentStratifiedKFold(n_splits=2)
        folds = list(kfold.split(Xa))[1]

        self.assertListEqual(list(folds[0]), [0, 1, 4, 5])
        self.assertListEqual(list(folds[1]), [2, 3, 6, 7])

    def test_multiclass_splits(self):
        X = pd.DataFrame({0: [10] * 12})  # Dummy corresponding X
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])
        y = pd.Series([0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1])
        kfold = TreatmentStratifiedKFold(n_splits=2)
        folds = list(kfold._split(X, a, y))[1]  # 2 folds, so first element suffices

        self.assertListEqual(
            list(folds[0]), [0, 1, 4, 5, 8, 9],
        )
        self.assertListEqual(
            list(folds[1]), [2, 3, 6, 7, 10, 11],
        )

    def test_against_sklearn_stratified_kfold(self):
        from sklearn.model_selection import StratifiedKFold

        X = pd.DataFrame({0: [10] * 4})
        a = pd.Series([0, 0, 1, 1])
        y = pd.Series([42.0, 42.1, 42.2, 42.3])

        kfold = TreatmentStratifiedKFold(n_splits=2, shuffle=False)
        folds = list(kfold._split(X, a, y))[0]  # 2 folds, so first element suffices

        sklearn_kfold = StratifiedKFold(n_splits=2, shuffle=False)
        sklearn_folds = list(sklearn_kfold.split(X, a))[0]  # 2 folds, so first element suffices
        self.assertListEqual(
            folds[0].tolist(), sklearn_folds[0].tolist()
        )
        self.assertListEqual(
            folds[1].tolist(), sklearn_folds[1].tolist()
        )

    def test_sklearn_search_compatibility(self):
        X = pd.DataFrame({0: [10] * 8})
        a = pd.Series([0, 0, 0, 0, 1, 1, 1, 1])
        y = pd.Series([0, 0, 1, 1, 0, 0, 1, 1])
        Xa = X.join(a.to_frame("a"))
        kfold = TreatmentStratifiedKFold(n_splits=2)
        folds = list(kfold.split(Xa, y))[1]

        self.assertListEqual(
            list(folds[0]), [0, 1, 4, 5],
        )
        self.assertListEqual(
            list(folds[1]), [2, 3, 6, 7],
        )



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jul 22, 2018

"""

import unittest
import random
import pandas as pd
import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.datasets import make_classification
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted

from causallib.preprocessing.confounder_selection import DoubleLASSO, RecursiveConfounderElimination


class _TestConfounderSelection(unittest.TestCase):
    def ensure_covariate_subset(self, selector, X, a, y, true_subset_confounders):
        # covariates_subset = ['x_0', 'x_2', f'x_{X.shape[1] - 1}', f'x_{X.shape[1] - 3}']
        covariates_subset = selector.covariates
        covariates_subset = list(set(covariates_subset) | set(true_subset_confounders))  # Ensure true-covariate in set

        selector.fit(X, a, y)
        X_trans = selector.transform(X)
        self.assertEqual(selector.n_features_, len(true_subset_confounders))
        for element in true_subset_confounders:
            self.assertIn(element, X_trans.columns)
        for element in set(covariates_subset) - set(true_subset_confounders):
            self.assertNotIn(element, X_trans.columns)
        self.assertGreater(X_trans.columns.difference(covariates_subset).size, 0)
        self.assertEqual(
            X_trans.shape[1],
            X.shape[1] - len(covariates_subset) + len(true_subset_confounders)
        )
        return selector

    def ensure_covariate_subset_binary(self, selector, X, a, y, true_subset_confounders):
        covariates_subset = selector.covariates | true_subset_confounders  # Ensure true-covariate in set

        selector.fit(X, a, y)
        X_trans = selector.transform(X)
        self.assertEqual(selector.n_features_, true_subset_confounders.sum())
        for element in X.columns[true_subset_confounders]:
            self.assertIn(element, X_trans.columns)
        for element in X.columns[covariates_subset].difference(X.columns[true_subset_confounders]):
            self.assertNotIn(element, X_trans.columns)
        self.assertGreater(X_trans.columns.difference(X.columns[covariates_subset]).size, 0)
        self.assertEqual(
            X_trans.shape[1],
            X.shape[1] - covariates_subset.sum() + true_subset_confounders.sum()
        )
        return selector


class TestDoubleLasso(_TestConfounderSelection):

    def setUp(self):
        # X is partitioned into two sets of xay_cols (parameter below)
        # First set is related to a and second set is related to y.
        # We generate each of this set using make_classification and
        # then merge them.
        self.xay_cols = 10
        self.min_accuracy = 0.95

    def make_xay(self, n_true_confounders_a, n_true_confounders_y, n_samples, seed=None):
        if seed is None:
            seed_a = random.randint(0, 2**16)
            seed_y = random.randint(0, 2**16)
        else:
            seed_a = seed
            seed_y = seed + 1
        xa, a = make_classification(n_samples=n_samples,
                                    n_features=self.xay_cols + 1,
                                    n_informative=min(n_true_confounders_a, self.xay_cols),
                                    n_redundant=0, n_repeated=0, class_sep=10.0,
                                    n_clusters_per_class=1,
                                    shuffle=False, random_state=seed_a)
        xy, y = make_classification(n_samples=n_samples,
                                    n_features=self.xay_cols + 1,
                                    n_informative=min(n_true_confounders_y, self.xay_cols),
                                    n_redundant=0, n_repeated=0, class_sep=10.0,
                                    n_clusters_per_class=1,
                                    shuffle=False, random_state=seed_y)
        x = np.concatenate((xa, xy), axis=1)
        x = StandardScaler().fit_transform(x)
        xdf = pd.DataFrame(x, columns=["x_" + str(i) for i in range(x.shape[1])])
        adf = pd.Series(a)
        ydf = pd.Series(y)
        return xdf, adf, ydf

    @staticmethod
    def make_simple_data():
        data = [[1, 1, 1, 1, 0, 0] * 5,     # x1 column: same as a
                [0, 0, 1, 0, 1, 1] * 5,     # x2 column: same as y
                [1, 2, 1, 2, 1, 2] * 5,     # x3 column: unrelated
                [1, 1, 1, 1, 0, 0] * 5,     # a (treatment)
                [0, 0, 1, 0, 1, 1] * 5,     # y
                ]
        col_names = ['x1', 'x2', 'x3', 'a', 'y']
        df = pd.DataFrame(list(zip(*data)), columns=col_names)
        X, a, y = df[['x1', 'x2', 'x3']], pd.Series(df['a']), pd.Series(df['y'])
        return X, a, y

    def test_correct_recovery_simple_data(self):
        X, a, y = self.make_simple_data()
        treatment_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        outcome_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        d = DoubleLASSO(treatment_lasso=treatment_lasso,
                        outcome_lasso=outcome_lasso,
                        threshold=0.1)
        d.fit(X, a, y)
        result = d.transform(X)
        treatment_cols = [X.columns[i] for i, c in enumerate(treatment_lasso.coef_.flatten())
                          if abs(c) > 1e-4]
        outcome_cols = [X.columns[i] for i, c in enumerate(outcome_lasso.coef_.flatten())
                        if abs(c) > 1e-4]
        self.assertEqual(len(treatment_cols), 1)
        self.assertEqual(len(outcome_cols), 1)
        self.assertListEqual(['x1', 'x2'], result.columns.tolist())

    @staticmethod
    def fit_transform(X, a, y, threshold=1e-6):
        treatment_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        outcome_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        dl = DoubleLASSO(treatment_lasso, outcome_lasso, threshold=threshold)
        dl.fit(X, a, y)
        result = dl.transform(X)
        score_a = treatment_lasso.score(X, a)
        score_y = outcome_lasso.score(X, y)
        return result, score_a, score_y

    def run_config(self, n_true_confounders_a, n_true_confounders_y, n_samples=100, threshold=1e-6):
        X, a, y = self.make_xay(n_true_confounders_a, n_true_confounders_y, n_samples)
        result, score_a, score_y = self.fit_transform(X, a, y, threshold)
        return result, score_a, score_y

    def test_basic(self):
        # Basic test by varying how many confounders are related to "a" and "y".
        # We also check for accuracy of treatment_lasso and outcome_lasso with
        # retained confounders as a proxy for how well we recover the confounders.
        for i in range(10):
            n_true_confounders_a = i + 1
            n_true_confounders_y = i + 1
            _, score_a, score_y = self.run_config(n_true_confounders_a, n_true_confounders_y)
            self.assertGreaterEqual(score_a, self.min_accuracy)
            self.assertGreaterEqual(score_y, self.min_accuracy)

    def test_threshold(self):
        # Vary zero-threshold and check that we select confounders as expected.
        n_true_confounders_a = 5
        n_true_confounders_y = 3
        for threshold in np.logspace(-6, 2, 9):     # Vary threshold from 1e-6 to 100 in powers of 10
            x, a, y = self.make_xay(n_true_confounders_a, n_true_confounders_y, n_samples=100)
            treatment_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
            outcome_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
            dl = DoubleLASSO(treatment_lasso, outcome_lasso, threshold=threshold)
            dl.fit(x, a, y)
            expected = ((np.abs(treatment_lasso.coef_.flatten()) >= threshold)
                        | (np.abs(outcome_lasso.coef_.flatten()) >= threshold))
            np.testing.assert_array_equal(dl.support_, expected)

    def test_mask_fn(self):

        # Custom merge function 1: True whenever array index modulo 3 is 0.
        def custom_mask_fn_and(tlasso, ylasso):
            t_coef = np.array([True if j % 3 != 2 else False
                               for j in range(len(tlasso.coef_.flatten()))], dtype=bool)
            y_coef = np.array([True if j % 3 != 1 else False
                               for j in range(len(ylasso.coef_.flatten()))], dtype=bool)
            return np.logical_and(t_coef, y_coef)

        # Custom merge function 2: True whenever array index modulo 3 is 0 or 1.
        def custom_mask_fn_or(tlasso, ylasso):
            t_coef = np.array([True if j % 3 == 0 else False
                               for j in range(len(tlasso.coef_.flatten()))], dtype=bool)
            y_coef = np.array([True if j % 3 == 1 else False
                               for j in range(len(ylasso.coef_.flatten()))], dtype=bool)
            return np.logical_or(t_coef, y_coef)

        # Vary number of confounders. Supply custom merge function.
        # Check at every step that the support_ attribute
        # of the fitted DoubleLasso is the same as expected from custom merge function.
        for i in range(10):
            n_true_confounders_a = i + 1
            n_true_confounders_y = i + 1
            X, a, y = self.make_xay(n_true_confounders_a, n_true_confounders_y, n_samples=100)
            treatment_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
            outcome_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)

            dl = DoubleLASSO(treatment_lasso, outcome_lasso, mask_fn=custom_mask_fn_and)
            dl.fit(X, a, y)
            np.testing.assert_array_equal(
                dl.support_, custom_mask_fn_and(treatment_lasso, outcome_lasso)
            )

            dl = DoubleLASSO(treatment_lasso, outcome_lasso, mask_fn=custom_mask_fn_or)
            dl.fit(X, a, y)
            np.testing.assert_array_equal(
                dl.support_, custom_mask_fn_or(treatment_lasso, outcome_lasso)
            )

    def test_covariate_subset(self):
        X, a, y = self.make_xay(2, 2, n_samples=100, seed=6)
        true_subset_confounders = ['x_0']  # Matches random seed: 6
        covariates_subset = ['x_0', 'x_2', f'x_{X.shape[1] - 1}', f'x_{X.shape[1] - 3}']

        treatment_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        outcome_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        dl = DoubleLASSO(
            treatment_lasso, outcome_lasso,
            covariates=covariates_subset,
        )

        dl = self.ensure_covariate_subset(dl, X, a, y, true_subset_confounders)

        np.testing.assert_array_equal(covariates_subset, dl.covariates)
        self.assertEqual(len(covariates_subset), dl.treatment_lasso.coef_.shape[1])
        self.assertEqual(len(covariates_subset), dl.outcome_lasso.coef_.shape[1])

    def test_covariate_subset_binary(self):
        X, a, y = self.make_xay(2, 2, n_samples=100, seed=6)
        true_subset_confounders = ['x_0']  # Matches random seed: 6
        covariates_subset = ['x_0', 'x_2', f'x_{X.shape[1] - 1}', f'x_{X.shape[1] - 3}']
        # Convert to binary:
        true_subset_confounders = X.columns.isin(true_subset_confounders)
        covariates_subset = X.columns.isin(covariates_subset)

        treatment_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        outcome_lasso = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        dl = DoubleLASSO(
            treatment_lasso, outcome_lasso,
            covariates=covariates_subset,
        )

        dl = self.ensure_covariate_subset_binary(dl, X, a, y, true_subset_confounders)

        np.testing.assert_array_equal(covariates_subset, dl.covariates)
        self.assertEqual(dl.covariates.sum(), dl.treatment_lasso.coef_.shape[1])
        self.assertEqual(dl.covariates.sum(), dl.outcome_lasso.coef_.shape[1])

    def test_auto_estimator_initialization(self):
        from sklearn.linear_model import LogisticRegressionCV, LassoCV

        X, a, y = self.make_xay(2, 2, n_samples=100, seed=6)
        y += np.random.normal(size=y.shape)  # Make `y` continuous to check auto-target-detection
        dl = DoubleLASSO()
        dl.fit(X, a, y)

        self.assertIsInstance(dl.treatment_lasso, LogisticRegressionCV)
        self.assertIsInstance(dl.outcome_lasso, LassoCV)

        self.assertIsNone(check_is_fitted(dl.treatment_lasso, "coef_"))
        self.assertIsNone(check_is_fitted(dl.outcome_lasso, "coef_"))

        self.assertEqual(dl.treatment_lasso.coef_.shape[1], X.shape[1])
        self.assertEqual(dl.outcome_lasso.coef_.shape[0], X.shape[1])  # Lasso different shape than LogisticRegression


class TestRecursiveConfounderElimination(_TestConfounderSelection):

    def setUp(self):
        self.max_x_cols = 10
        self.min_accuracy = 0.95

    def make_xay(self, n_true_confounders, n_samples, seed=None):
        xa, y = make_classification(n_samples=n_samples,
                                    n_features=self.max_x_cols + 1,
                                    n_informative=min(n_true_confounders, self.max_x_cols),
                                    n_redundant=0, n_repeated=0, class_sep=10.0,
                                    n_clusters_per_class=1,
                                    shuffle=False, random_state=seed)
        xa = StandardScaler().fit_transform(xa)
        a, x = xa[:, -1], xa[:, :-1]
        xdf = pd.DataFrame(x, columns=["x_" + str(i) for i in range(x.shape[1])])
        adf = pd.Series(a)
        ydf = pd.Series(y)
        return xdf, adf, ydf

    @staticmethod
    def fit_transform(X, a, y, n_true_confounders, step=1):
        estimator = LogisticRegression(penalty='l1', solver='saga', max_iter=5000)
        rfe = RecursiveConfounderElimination(estimator=estimator,
                                             n_features_to_select=n_true_confounders,
                                             step=step)
        rfe.fit(X, a, y)
        score = estimator.score(a.to_frame().join(X.iloc[:, rfe.support_]), y)
        result = rfe.transform(X)
        return result, score

    def run_config(self, n_true_confounders, n_samples=100, step=1):
        X, a, y = self.make_xay(n_true_confounders, n_samples)
        result, score = self.fit_transform(X, a, y, n_true_confounders, step)
        return result, score

    def test_basic(self):
        # Basic test with increasing number of confounders.
        # We check that the required number of confounders are returned.
        # We also check for accuracy of classifier with retained confounders
        # as a proxy for how well we recover the confounders.
        for i in range(self.max_x_cols):
            n_true_confounders = i + 1
            result, score = self.run_config(n_true_confounders=n_true_confounders)
            self.assertEqual(len(result.columns), n_true_confounders)
            self.assertGreaterEqual(score, self.min_accuracy)

    def test_more_features_than_confounders(self):
        # Test for n_features_to_select > X.columns: in this case
        # we should return the entire X as answer.
        # We also check for accuracy of classifier with retained confounders
        # as a proxy for how well we recover the confounders.
        for i in range(10):
            n_true_confounders = self.max_x_cols + i + 1
            result, score = self.run_config(n_true_confounders=n_true_confounders)
            self.assertEqual(len(result.columns), self.max_x_cols)
            self.assertGreaterEqual(score, self.min_accuracy)

    def test_step(self):
        # Test for different step sizes (i.e., confounders to eliminate in each step).
        # We also check for accuracy of classifier with retained confounders
        # as a proxy for how well we recover the confounders.
        n_true_confounders = 3
        for i in range(self.max_x_cols + 5):    # test from step = 1 to > max columns
            step = i + 1
            result, score = self.run_config(n_true_confounders=n_true_confounders, step=step)
            self.assertEqual(len(result.columns), n_true_confounders)
            self.assertGreaterEqual(score, self.min_accuracy)

    def test_covariates_subset(self):
        X, a, y = self.make_xay(2, n_samples=100, seed=6)
        true_subset_confounders = ['x_0']  # Matches random seed: 6
        covariates_subset = ['x_0', 'x_2', f'x_{X.shape[1] - 1}', f'x_{X.shape[1] - 3}']

        rfe = RecursiveConfounderElimination(
            estimator=LogisticRegression(),
            covariates=covariates_subset,
        )

        rfe = self.ensure_covariate_subset(rfe, X, a, y, true_subset_confounders)

        np.testing.assert_array_equal(covariates_subset, rfe.covariates)
        self.assertEqual(len(true_subset_confounders), rfe.n_features_)

    def test_covariate_subset_binary(self):
        X, a, y = self.make_xay(2, n_samples=100, seed=6)
        true_subset_confounders = ['x_0']  # Matches random seed: 6
        covariates_subset = ['x_0', 'x_2', f'x_{X.shape[1] - 1}', f'x_{X.shape[1] - 3}']
        # Convert to binary:
        true_subset_confounders = X.columns.isin(true_subset_confounders)
        covariates_subset = X.columns.isin(covariates_subset)

        rfe = RecursiveConfounderElimination(
            estimator=LogisticRegression(),
            covariates=covariates_subset,
        )

        rfe = self.ensure_covariate_subset_binary(rfe, X, a, y, true_subset_confounders)

        np.testing.assert_array_equal(covariates_subset, rfe.covariates)
        self.assertEqual(true_subset_confounders.sum(), rfe.n_features_)



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Sep 9, 2021

"""

import unittest
import itertools
import operator
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression, LinearRegression
from causallib.estimation import StratifiedStandardization, XLearner
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.datasets import make_classification
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError


class TestXLearner(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        X, y = make_classification(n_samples=500,
                                   n_features=3,
                                   n_informative=2,
                                   n_redundant=0,
                                   n_repeated=0,
                                   n_classes=2,
                                   n_clusters_per_class=1,
                                   flip_y=0.01,
                                   class_sep=1.0)
        X, a = X[:, :-1], X[:, -1]
        # so at least a quarter of the samples will be positive or negative
        q = np.random.rand() * 1 / 2 + 1 / 4
        # the following is a random treatment assignment
        a = (a > np.percentile(a, q * 100)).astype(int)
        sum_x = np.sum(X[:, :-1], axis=1)
        # The following is a dependent treatment assignment
        dependent_a = sum_x + np.random.rand() * np.max(sum_x) * 0.1
        dependent_a = (dependent_a > np.percentile(dependent_a, q * 100)).astype(int)
        cls.data_2cls = {"X": pd.DataFrame(X),
                         "a": pd.Series(a),
                         "dependent_a": pd.Series(dependent_a),
                         "y": pd.Series(y),
                         "quantile": q}

    def setUp(self):
        self.estimator = XLearner(
            outcome_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_model=StratifiedStandardization(LinearRegression(normalize=True))
        )

    def default_fit(self):
        self.estimator.fit(self.data_2cls["X"], self.data_2cls["a"], self.data_2cls["y"])

    def test_dummy(self):
        self.default_fit()
        pred = self.estimator.treatment_model.predict_proba(self.data_2cls["X"])
        expect = np.repeat(np.mean(self.data_2cls["a"]), self.data_2cls["X"].shape[0])
        np.testing.assert_array_almost_equal(pred[:, 1], expect)
        np.testing.assert_array_almost_equal(pred[:, 0], 1 - expect)

    def fit_logistic_treatment_model(self, effect_types='diff'):
        estimator = XLearner(
            outcome_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_model=StratifiedStandardization(LinearRegression(normalize=True)),
            treatment_model=LogisticRegression(),
            effect_types=effect_types
        )
        estimator.fit(self.data_2cls["X"], self.data_2cls["a"], self.data_2cls["y"])
        return estimator

    def obtain_dummy_treat_model(self, effect_types='diff'):
        return XLearner(
            outcome_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_types=effect_types
        )

    def test_logistic_fit(self):
        estimator = self.fit_logistic_treatment_model()
        self.assertIsInstance(estimator.treatment_model, LogisticRegression)
        self.assertIsNone(check_is_fitted(estimator.treatment_model, attributes="coef_"))

    def test_logistic_treatment_values(self):
        estimator = self.fit_logistic_treatment_model()
        pred = estimator.treatment_model.predict_proba(self.data_2cls["X"])
        new_model = LogisticRegression().fit(self.data_2cls["X"], self.data_2cls["a"])
        expect = new_model.predict_proba(self.data_2cls["X"])
        np.testing.assert_array_almost_equal(pred, expect)

    def test_outcome_fit(self):
        self.default_fit()
        # Following method throws an exeption if not fitted otherwise returns none
        self.assertIsNone(check_is_fitted(self.estimator.outcome_model.learner[0], attributes="coef_"))
        self.assertIsNone(check_is_fitted(self.estimator.outcome_model.learner[1], attributes="coef_"))

    def test_cate_fit(self):
        self.default_fit()
        # Following method throws an exeption if not fitted otherwise returns none
        self.assertIsNone(check_is_fitted(self.estimator.effect_model.learner[0], attributes="coef_"))
        self.assertIsNone(check_is_fitted(self.estimator.effect_model.learner[1], attributes="coef_"))

    def test_dummy_fit(self):
        self.default_fit()
        # The dummy model does not have an indicator attribute
        try:
            self.estimator.treatment_model.predict([1])
        except NotFittedError:
            self.assertTrue(False)
        self.assertTrue(True)

    def test_population_estimation(self):
        self.default_fit()
        eff = self.estimator.estimate_effect(
            self.data_2cls["X"], self.data_2cls["a"], agg="population"
        )
        self.assertEqual(len(eff), 1)

    def test_individual_estimation(self):
        self.default_fit()
        eff = self.estimator.estimate_effect(
            self.data_2cls["X"],
            self.data_2cls["a"],
            agg="individual"
        )
        self.assertEqual(len(eff), len(self.data_2cls["a"]))

    def test_fixed_effect_diff(self):
        est = self.obtain_dummy_treat_model(effect_types='diff')
        est.fit(
            self.data_2cls["X"],
            self.data_2cls["a"],
            self.data_2cls["a"] * 2
        )
        eff = est.estimate_effect(
            self.data_2cls["X"],
            self.data_2cls["a"],
            agg="individual"
        )
        self.assertEqual(np.mean(eff.values), 2)

    def test_fixed_effect_ratio(self):
        est = self.obtain_dummy_treat_model(effect_types='ratio')
        est.fit(
            self.data_2cls["X"],
            self.data_2cls["a"],
            self.data_2cls["a"] * 2 + 1
        )
        eff = est.estimate_effect(
            self.data_2cls["X"],
            self.data_2cls["a"],
            agg="individual"
        )
        self.assertAlmostEqual(np.mean(eff.values), 3)

    def test_fixed_effect_or(self):
        est = self.obtain_dummy_treat_model(effect_types='or')
        est.fit(
            self.data_2cls["X"],
            self.data_2cls["a"],
            self.data_2cls["a"] * 2 + 2
        )
        eff = est.estimate_effect(
            self.data_2cls["X"],
            self.data_2cls["a"],
            agg="individual"
        )
        self.assertAlmostEqual(np.mean(eff.values), 2 / 3)

    def test_estimate_effect(self):
        self.default_fit()
        eff = self.estimator.estimate_effect(
            self.data_2cls["X"],
            self.data_2cls["a"]
        )
        self.assertEqual(np.mean(np.isnan(eff)), 0)

    def _obtain_zero_effect_set(self):
        big_X = pd.DataFrame(
            np.concatenate((self.data_2cls["X"], self.data_2cls["X"]), axis=0)
        )
        big_y = pd.Series(
            np.concatenate((self.data_2cls["y"], self.data_2cls["y"]))
        )
        big_a = pd.Series(
            np.concatenate((np.ones(len(self.data_2cls["y"])), np.zeros(len(self.data_2cls["y"]))))
        )
        return big_X, big_a, big_y

    def test_zero_effect(self):
        big_X, big_a, big_y = self._obtain_zero_effect_set()
        self.estimator.fit(big_X, big_a, big_y)
        eff = self.estimator.estimate_effect(big_X, big_a)
        self.assertAlmostEqual(eff.values, 0, places=5)

    def test_zero_effect_forests(self):
        big_X, big_a, big_y = self._obtain_zero_effect_set()
        estimator_forest = XLearner(
            outcome_model=StratifiedStandardization(RandomForestClassifier()),
            effect_model=StratifiedStandardization(RandomForestRegressor()),
        )
        estimator_forest.fit(big_X, big_a, big_y)
        eff = estimator_forest.estimate_effect(big_X, big_a)
        np.testing.assert_array_almost_equal(eff.values, 0, decimal=2) # Random forest is more noisy

    @staticmethod
    def _obtain_simple_set():
        y_pred = pd.DataFrame({
            0: [1, 2, 3, 4],
            1: [10, 20, 30, 40],
        })
        y_true = pd.Series([100, 200, 300, 400])
        a = pd.Series([0, 0, 1, 1])
        return y_pred, y_true, a

    def test_diff_imputed_treatment_effect(self):
        y_pred, y_true, a = self._obtain_simple_set()
        imp_effect = self.estimator._obtain_imputed_treatment_effect(
            y_pred, a, y_true,
        )
        expected = pd.Series([
            10 - 100,  # y1 - mu0(X1)
            20 - 200,  # y1 - mu0(X1)
            300 - 3,  # mu1(X0) - y0
            400 - 4,  # mu1(X0) - y0
        ])
        pd.testing.assert_series_equal(imp_effect, expected)

    def test_ratio_imputed_treatment_effect(self):
        y_pred, y_true, a = self._obtain_simple_set()
        est = self.obtain_dummy_treat_model(effect_types='ratio')
        imp_effect = est._obtain_imputed_treatment_effect(
            y_pred, a, y_true
        )
        expected = pd.Series([
            10 / 100,  # y1 - mu0(X1)
            20 / 200,  # y1 - mu0(X1)
            300 / 3,  # mu1(X0) - y0
            400 / 4,  # mu1(X0) - y0
        ])
        pd.testing.assert_series_equal(imp_effect, expected)

    def test_or_imputed_treatment_effect(self):
        y_pred, y_true, a = self._obtain_simple_set()
        est = self.obtain_dummy_treat_model(effect_types='or')
        imp_effect = est._obtain_imputed_treatment_effect(
            y_pred, a, y_true
        )
        or_func = self.estimator.CALCULATE_EFFECT['or']
        expected = pd.Series([
            or_func(10, 100),  # y1 - mu0(X1)
            or_func(20, 200),  # y1 - mu0(X1)
            or_func(300, 3),  # mu1(X0) - y0
            or_func(400, 4)  # mu1(X0) - y0
        ])
        pd.testing.assert_series_equal(imp_effect, expected)

    def test_individual_outcome_warning(self):
        self.default_fit()

        self.assertWarns(
            Warning,
            self.estimator.estimate_individual_outcome, # The method to test for warning
            self.data_2cls['X'], self.data_2cls['a'] # The arguments for assertWarns are passed to the tested method
        )

    def test_individual_outcome_xlearner_outcome_model(self):
        self.default_fit()
        by_xlearner = self.estimator.estimate_individual_outcome(self.data_2cls['X'], self.data_2cls['a'])
        by_outcome = self.estimator.outcome_model.estimate_individual_outcome(self.data_2cls['X'], self.data_2cls['a'])
        np.testing.assert_array_almost_equal(by_xlearner, by_outcome)

    def _obtain_dual_indivdual_estimation(self, effect_types_first='diff', effect_types_second='diff'):
        estimator_first = XLearner(
            outcome_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_types=effect_types_first
        )
        estimator_second = XLearner(
            outcome_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_model=StratifiedStandardization(LinearRegression(normalize=True)),
            effect_types=effect_types_second
        )
        mod_y = self.data_2cls['y'] + 2 # the plus two is for the ratio so we will divide by zero
        estimator_first.fit(self.data_2cls['X'], self.data_2cls['a'], mod_y)
        estimator_second.fit(self.data_2cls['X'], self.data_2cls['a'], mod_y)
        est_ef_first = estimator_first.estimate_effect(self.data_2cls['X'], self.data_2cls['a'])
        est_ef_second = estimator_second.estimate_effect(self.data_2cls['X'], self.data_2cls['a'])
        return est_ef_first, est_ef_second

    def test_same_effect_type_diff(self):
        for eff in self.estimator.CALCULATE_EFFECT.keys():
            np.testing.assert_array_almost_equal(
                *self._obtain_dual_indivdual_estimation(effect_types_first=eff, effect_types_second=eff)
            )

    def test_different_effect_type_diff(self):
        eff_list = self.estimator.CALCULATE_EFFECT.keys()
        for eff_frst, eff_scnd in itertools.permutations(eff_list, 2):
            est_ef_fit, est_ef_init = self._obtain_dual_indivdual_estimation(
                effect_types_first=eff_frst, effect_types_second=eff_scnd)
            # numpy does not have assert array not equal
            np.testing.assert_array_compare(operator.__ne__, est_ef_fit, est_ef_init)

    def test_deep_copy_at_fit(self):
        estimator = XLearner(outcome_model=StratifiedStandardization(LinearRegression()))
        self.assertIsNone(estimator.effect_model)
        estimator.fit(self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])
        self.assertIsInstance(estimator.effect_model, StratifiedStandardization)
        self.assertIsNot(estimator.effect_model, estimator.outcome_model)

    def test_predict_proba_miss_match(self):
        estimator_miss_match = XLearner(
            outcome_model=StratifiedStandardization(LogisticRegression(), predict_proba=False),
            effect_model=StratifiedStandardization(LinearRegression(), predict_proba=True),
            # following is the default value it is written to make it more explicit
            predict_proba=True,
        )
        estimator_miss_match.fit(self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])

        estimator_match = XLearner(
            outcome_model=StratifiedStandardization(LogisticRegression(), predict_proba=True),
            effect_model=StratifiedStandardization(LinearRegression(), predict_proba=True),
            # following is the default value it is written to make it more explicit
            predict_proba=True,
        )
        estimator_match.fit(self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])
        miss_match_est = estimator_miss_match.estimate_effect(self.data_2cls['X'], self.data_2cls['a'])
        match_est = estimator_match.estimate_effect(self.data_2cls['X'], self.data_2cls['a'])
        np.testing.assert_array_almost_equal(miss_match_est, match_est)

    def test_predict_proba_False(self):
        estimator_miss_match = XLearner(
            outcome_model=StratifiedStandardization(LogisticRegression(), predict_proba=False),
            effect_model=StratifiedStandardization(LinearRegression()),
            # following is the default value it is written to make it more explicit
            predict_proba=False,
        )
        estimator_miss_match.fit(
            self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])
        outcomes = estimator_miss_match._obtain_ind_outcomes(
            self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])
        np.testing.assert_array_equal(np.sort(np.unique(outcomes)), np.array([0, 1]))

    def test_predict_proba_false_overide(self):
        estimator_miss_match = XLearner(
            outcome_model=StratifiedStandardization(LogisticRegression(), predict_proba=False),
            effect_model=StratifiedStandardization(LinearRegression()),
            # following is the default value it is written to make it more explicit
            predict_proba=True,
        )
        estimator_miss_match.fit(
            self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])
        outcomes = estimator_miss_match._obtain_ind_outcomes(
            self.data_2cls['X'], self.data_2cls['a'], self.data_2cls['y'])
        self.assertFalse(self._is_binary_values_after_rounding(outcomes))

    def test_is_binary_values_after_rounding(self):
        sample_size = 3000
        decimals_number = 5
        decimal_fraction = 10**(-decimals_number)
        binary_vals = np.random.choice([0, 1], size=sample_size)
        noise_values = np.random.rand(sample_size)*decimal_fraction
        to_test_vec = binary_vals+noise_values
        for decimal_test in np.arange(1, decimals_number*2):
            if decimal_test >= decimals_number:
                self.assertFalse(
                    self._is_binary_values_after_rounding(to_test_vec, digit=decimal_test))
            else:
                self.assertTrue(
                    self._is_binary_values_after_rounding(to_test_vec, digit=decimal_test))


    @staticmethod
    def _is_binary_values_after_rounding(vec, digit=3):
        un = np.sort(np.unique(np.round(vec, digit)))
        return np.array_equal(un, [0, 1])

    def test_init_effect(self):
        for effect in ['diff', 'ratio', 'or']:
            est = self.obtain_dummy_treat_model(effect_types=effect)
            self.assertEqual(est.effect_types, effect)

    def test_effect_type_name_population(self):
        agg = "population"
        for effect in ['ratio', 'or', 'diff']:
            est = self.obtain_dummy_treat_model(effect_types=effect)
            est.fit(
                self.data_2cls["X"],
                self.data_2cls["a"],
                self.data_2cls["a"] * 2 + 1
            )
            eff = est.estimate_effect(
                self.data_2cls["X"],
                self.data_2cls["a"],
                agg=agg
            )
        self.assertEqual(eff.index[0], effect)

    def test_effect_type_name_individual(self):
        agg = "individual"
        for effect in ['ratio', 'or', 'diff']:
            est = self.obtain_dummy_treat_model(effect_types=effect)
            est.fit(
                self.data_2cls["X"],
                self.data_2cls["a"],
                self.data_2cls["a"] * 2 + 1
            )
            eff = est.estimate_effect(
                self.data_2cls["X"],
                self.data_2cls["a"],
                agg=agg
            )
            self.assertEqual(eff.name, effect)


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jun 09, 2021

"""

import unittest
# from unittest import TestCase

# import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from causallib.estimation.overlap_weights import OverlapWeights


class TestOverlapWeights(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Data:
        X, a = make_classification(n_features=1, n_informative=1, n_redundant=0, n_repeated=0, n_classes=2,
                                   n_clusters_per_class=1, flip_y=0.0, class_sep=10.0)
        cls.data_r_100 = {"X": pd.DataFrame(X), "a": pd.Series(a)}
        X, a = make_classification(n_features=1, n_informative=1, n_redundant=0, n_repeated=0, n_classes=2,
                                   n_clusters_per_class=1, flip_y=0.2, class_sep=10.0)
        cls.data_r_80 = {"X": pd.DataFrame(X), "a": pd.Series(a)}

        # Data that maps x=0->a=0 and x=1->a=1:
        X = pd.Series([0] * 50 + [1] * 50)
        cls.data_cat_r_100 = {"X": X.to_frame(), "a": X}

        # Data that maps x=0->a=0 and x=1->a=1, but 10% of x=0->a=1 and 10% of x=1->a=0:
        X = pd.Series([0] * 40 + [1] * 10 + [1] * 40 + [0] * 10).to_frame()
        a = pd.Series([0] * 50 + [1] * 50)
        cls.data_cat_r_80 = {"X": X, "a": a}

        # Avoids regularization of the model:
        cls.estimator = OverlapWeights(LogisticRegression(C=1e6, solver='lbfgs'), use_stabilized=False)

    def setUp(self):
        self.estimator.fit(self.data_r_100["X"], self.data_r_100["a"])

    def test_classes_number_is_two(self):
        with self.subTest("OW check error arise if single class"):
            a = pd.Series(0, index=self.data_r_100["X"].index)
            with self.assertRaises(AssertionError):
                self.estimator.compute_weight_matrix(self.data_r_100["X"], a)
        with self.subTest("OW check error arise if more than two classes"):
            a = pd.Series([0] * 30 + [1] * 30 + [2] * 40, index=self.data_r_100["X"].index)
            with self.assertRaises(AssertionError):
                self.estimator.compute_weight_matrix(self.data_r_100["X"], a)

    def test_truncate_values_not_none(self):
        with self.assertWarns(RuntimeWarning):
            self.estimator.compute_weight_matrix(
                self.data_r_100["X"], self.data_r_100["a"],
                clip_min=0.2, clip_max=0.8, use_stabilized=None)

    def test_categorical_classes_df_col_names(self):
        a = pd.Series(["a"] * 50 + ["b"] * 50, index=self.data_r_100["X"].index)
        w = self.estimator.compute_weight_matrix(self.data_r_100["X"], a)
        cols_w = w.columns.values.tolist()
        self.assertTrue(cols_w, ["a", "b"])

    def test_ow_weights_reversed_to_propensity(self):
        propensity = self.estimator.learner.predict_proba(self.data_r_100["X"])
        propensity = pd.DataFrame(propensity)
        ow_weights = self.estimator.compute_weight_matrix(self.data_r_100["X"], self.data_r_100["a"],
                                                          clip_min=None, clip_max=None)
        pd.testing.assert_series_equal(propensity.loc[:, 0], ow_weights.loc[:, 1], check_names=False)
        pd.testing.assert_series_equal(propensity.loc[:, 1], ow_weights.loc[:, 0], check_names=False)
        pd.testing.assert_index_equal(propensity.columns, ow_weights.columns)


"""Plot data extractors.

The responsibility of these classes is to extract the data from the
EvaluationResults objects to match the requested plot.
"""

import abc

from sklearn import metrics

from . import curve_data_makers, plots
from ...metrics.weight_metrics import calculate_covariate_balance


class BaseEvaluationPlotDataExtractor(abc.ABC):
    """Extractor to get plot data from EvaluationResults.

    Subclasses also have a `plot_names` property.
    """

    def __init__(
        self, evaluation_results: "causallib.evaluation.results.EvaluationResults"
    ):
        self.predictions = evaluation_results.predictions
        self.X = evaluation_results.X
        self.a = evaluation_results.a
        self.y = evaluation_results.y
        self.cv = evaluation_results.cv

    def cv_by_phase(self, phase="train"):
        """Get the cross-validation indices of all folds for a given phase.

        Args:
            phase (str, optional): Requested phase: "train" or "valid. Defaults to "train".

        Returns:
            List: _description_
        """
        fold_idx = 0 if phase == "train" else 1
        return [fold[fold_idx] for fold in self.cv]

    @abc.abstractmethod
    def get_data_for_plot(self, plot_name, phase="train"):
        """Get data for plot with name `plot_name`."""
        raise NotImplementedError


class WeightPlotDataExtractor(BaseEvaluationPlotDataExtractor):
    """Extractor to get plot data from WeightEvaluatorPredictions."""

    plot_names = plots.WeightPlotNames

    def get_data_for_plot(self, plot_name, phase="train"):
        """Retrieve the data needed for each provided plot.

        Plot functions are in plots module.

        Args:
            plot_name (str): Plot name.

        Returns:
            tuple: Plot data
        """

        folds_predictions = self.predictions[phase]
        if plot_name in {plots.WEIGHT_DISTRIBUTION_PLOT}:
            return (
                [p.weight_for_being_treated for p in folds_predictions],
                self.a,
                self.cv_by_phase(phase),
            )
        if plot_name in {
            plots.COVARIATE_BALANCE_LOVE_PLOT,
            plots.COVARIATE_BALANCE_SLOPE_PLOT,
            plots.COVARIATE_BALANCE_GENERIC_PLOT,
        }:
            distribution_distances = []
            for fold_prediction in folds_predictions:
                fold_w = fold_prediction.weight_by_treatment_assignment
                fold_X = self.X.loc[fold_w.index]
                fold_a = self.a.loc[fold_w.index]
                dist_dist = calculate_covariate_balance(fold_X, fold_a, fold_w)
                distribution_distances.append(dist_dist)
            return (distribution_distances,)

        raise ValueError(f"Received unsupported plot name {plot_name}!")


class PropensityPlotDataExtractor(WeightPlotDataExtractor):
    """Extractor to get plot data from PropensityEvaluatorPredictions."""

    plot_names = plots.PropensityPlotNames

    def get_data_for_plot(self, plot_name, phase="train"):
        """Retrieve the data needed for each provided plot.
        Plot interfaces are at the plots.py module.

        Args:
            plot_name (str): Plot name.
            fold_predictions (list[PropensityEvaluatorPredictions]): Predictions for each fold.
            cv list[np.ndarray]: Indices (in iloc positions) of each fold.

        Returns:
            tuple: Plot data
        """
        fold_predictions = self.predictions[phase]

        if plot_name in {plots.WEIGHT_DISTRIBUTION_PLOT, plots.CALIBRATION_PLOT}:
            return (
                [p.propensity for p in fold_predictions],
                self.a,
                self.cv_by_phase(phase),
            )
        if plot_name in {plots.ROC_CURVE_PLOT}:
            curve_data = curve_data_makers.calculate_curve_data_propensity(
                fold_predictions, self.a, metrics.roc_curve, metrics.roc_auc_score
            )
            roc_curve = curve_data_makers.calculate_roc_curve(curve_data)
            return (roc_curve,)
        if plot_name in {plots.PR_CURVE_PLOT}:
            curve_data = curve_data_makers.calculate_curve_data_propensity(
                fold_predictions,
                self.a,
                metrics.precision_recall_curve,
                metrics.average_precision_score,
            )
            pr_curve = curve_data_makers.calculate_pr_curve(curve_data, self.a)
            return (pr_curve,)

        # Common plots are implemented at top-most level possible.
        # Plot might be implemented by WeightEvaluator:
        return super().get_data_for_plot(plot_name, phase=phase)


class ContinuousOutcomePlotDataExtractor(BaseEvaluationPlotDataExtractor):
    """Extractor to get plot data from OutcomeEvaluatorPredictions.

    Note that the available plots are different if the outcome predictions
    are binary/classification or continuous/regression.
    """

    plot_names = plots.ContinuousOutputPlotNames

    def get_data_for_plot(self, plot_name, phase="train"):
        """Retrieve the data needed for each provided plot.
        Plot interfaces are at the plots module.

        Args:
            plot_name (str): Plot name.

        Returns:
            tuple: Plot data
        """
        fold_predictions = self.predictions[phase]
        if plot_name in {plots.CONTINUOUS_ACCURACY_PLOT, plots.RESIDUALS_PLOT}:
            return (
                [x.get_prediction_by_treatment(self.a) for x in fold_predictions],
                self.y,
                self.a,
                self.cv_by_phase(phase),
            )
        if plot_name in {plots.COMMON_SUPPORT_PLOT}:
            return (
                [p.prediction for p in fold_predictions],
                self.a,
                self.cv_by_phase(phase),
            )

        raise ValueError(f"Received unsupported plot name {plot_name}!")


class BinaryOutcomePlotDataExtractor(BaseEvaluationPlotDataExtractor):
    """Extractor to get plot data from OutcomeEvaluatorPredictions.

    Note that the available plots are different if the outcome predictions
    are binary/classification or continuous/regression.
    """

    plot_names = plots.BinaryOutputPlotNames

    def get_data_for_plot(self, plot_name, phase="train"):
        """Retrieve the data needed for each provided plot.
        Plot interfaces are at the plots module.

        Args:
            plot_name (str): Plot name.

        Returns:
            tuple: Plot data
        """
        fold_predictions = self.predictions[phase]

        if plot_name in {plots.CALIBRATION_PLOT}:
            return (
                [x.get_proba_by_treatment(self.a) for x in fold_predictions],
                self.y,
                self.cv_by_phase(phase),
            )
        if plot_name in {plots.ROC_CURVE_PLOT}:
            proba_list = [x.get_proba_by_treatment(self.a) for x in fold_predictions]
            curve_data = curve_data_makers.calculate_curve_data_binary_outcome(
                proba_list,
                self.y,
                metrics.roc_curve,
                metrics.roc_auc_score,
                stratify_by=self.a,
            )
            roc_curve_data = curve_data_makers.calculate_roc_curve(curve_data)
            return (roc_curve_data,)

        if plot_name in {plots.PR_CURVE_PLOT}:
            proba_list = [x.get_proba_by_treatment(self.a) for x in fold_predictions]
            curve_data = curve_data_makers.calculate_curve_data_binary_outcome(
                proba_list,
                self.y,
                metrics.precision_recall_curve,
                metrics.average_precision_score,
                stratify_by=self.a,
            )
            pr_curve_data = curve_data_makers.calculate_pr_curve(curve_data, self.y)
            return (pr_curve_data,)

        raise ValueError(f"Received unsupported plot name {plot_name}!")



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Aug 29, 2018

"""
import abc
import unittest

import pandas as pd
import numpy as np

from causallib.preprocessing.transformers import StandardScaler, MinMaxScaler


class TestTransformers:
    @abc.abstractmethod
    def test_fit(self):
        pass

    @abc.abstractmethod
    def test_fit_technical(self):
        pass

    @abc.abstractmethod
    def test_transform(self):
        pass

    @abc.abstractmethod
    def test_inverse_transform(self):
        pass


class TestStandardScaler(TestTransformers, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = pd.DataFrame({"binary": [4, 5, 5, np.nan],
                                 "continuous": [0, 2, 4, np.nan]})
        cls.transformer = StandardScaler(with_mean=True, with_std=True, ignore_nans=True)
        cls.transformer.fit(cls.data)

    def test_fit_technical(self):
        with self.subTest("Has mean_ attribute"):
            self.assertTrue(hasattr(self.transformer, "mean_"))

        with self.subTest("Has scale_ attribute"):
            self.assertTrue(hasattr(self.transformer, "scale_"))

        with self.subTest("Applied on the right amount of columns"):
            # Should only be applied on "continuous" column
            self.assertEqual(1, len(self.transformer.mean_))
            self.assertEqual(1, len(self.transformer.scale_))

    def test_fit(self):
        with self.subTest("Test means are right"):
            self.assertEqual(2.0, self.transformer.mean_["continuous"])

        with self.subTest("Test scale is correct"):
            self.assertEqual(2.0, self.transformer.scale_["continuous"])

    def test_transform(self):
        transformed = self.transformer.transform(self.data)

        with self.subTest("Was not applied on binary column"):
            pd.testing.assert_series_equal(self.data["binary"], transformed["binary"])

        with self.subTest("Result is right on the transformed column"):
            pd.testing.assert_series_equal(transformed["continuous"], pd.Series([-1.0, 0.0, 1.0, np.nan]),
                                           check_names=False)

    def test_inverse_transform(self):
        untransformed = self.transformer.inverse_transform(self.transformer.transform(self.data))
        pd.testing.assert_frame_equal(self.data, untransformed)


class TestMinMaxScaler(TestTransformers, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = pd.DataFrame({"binary": [4, 5, 5, np.nan],
                                 "continuous": [0, 2, 4, np.nan]})
        cls.transformer = MinMaxScaler(only_binary_features=True, ignore_nans=True)
        cls.transformer.fit(cls.data)

    def test_fit_technical(self):
        with self.subTest("Has min_ attribute"):
            self.assertTrue(hasattr(self.transformer, "min_"))

        with self.subTest("Has max_ attribute"):
            self.assertTrue(hasattr(self.transformer, "max_"))

        with self.subTest("Has scale_ attribute"):
            self.assertTrue(hasattr(self.transformer, "scale_"))

        with self.subTest("Applied on the right amount of columns"):
            # Should only be applied on "continuous" column
            self.assertEqual(1, len(self.transformer.min_))
            self.assertEqual(1, len(self.transformer.max_))
            self.assertEqual(1, len(self.transformer.scale_))

    def test_fit(self):
        with self.subTest("Test min is right"):
            self.assertEqual(4.0, self.transformer.min_["binary"])

        with self.subTest("Test max is right"):
            self.assertEqual(5.0, self.transformer.max_["binary"])

        with self.subTest("Test scale is correct"):
            self.assertEqual(1.0, self.transformer.scale_["binary"])

    def test_transform(self):
        transformed = self.transformer.transform(self.data)

        with self.subTest("Was not applied on binary column"):
            pd.testing.assert_series_equal(self.data["continuous"], transformed["continuous"])

        with self.subTest("Result is right on the transformed column"):
            pd.testing.assert_series_equal(transformed["binary"], pd.Series([0.0, 1.0, 1.0, np.nan]),
                                           check_names=False)

    def test_inverse_transform(self):
        untransformed = self.transformer.inverse_transform(self.transformer.transform(self.data))
        pd.testing.assert_frame_equal(self.data, untransformed)



import numpy as np
import pandas as pd
import torch
import unittest

from causallib.contrib.hemm import HEMM
from causallib.contrib.hemm.hemm_utilities import genSplits, returnIndices, getMeanandStd
from causallib.contrib.hemm.hemm_outcome_models import genMLPModule, genLinearModule, BalancedNet
from causallib.contrib.hemm.gen_synthetic_data import gen_montecarlo


class TestHemmEstimator(unittest.TestCase):
    def experiment(self, data, i, comp, response, outcome_model, lr, batch_size):
        np.random.seed(0)

        Xtr = data['TRAIN']['x'][:, :, i]
        Ttr = data['TRAIN']['t'][:, i]
        Ytr = data['TRAIN']['yf'][:, i]

        Ytr_ = np.ones_like(Ytr)
        splits = genSplits(Ttr, Ytr_)
        train, dev = returnIndices(splits)
        n = Xtr.shape[0]

        Xte = data['TEST']['x'][:, :, i]
        Tte = data['TEST']['t'][:, i]
        Yte = data['TEST']['yf'][:, i]

        mu, std = getMeanandStd(Xtr)

        Xdev = Xtr[dev]  # Numpy array
        Ydev = torch.from_numpy(Ytr[dev].astype('float64'))
        Tdev = Ttr[dev]  # Numpy array

        Xtr = pd.DataFrame(Xtr[train])  # Train covariates as a data frame
        Ytr = torch.from_numpy(Ytr[train].astype('float64'))
        Ttr = pd.Series(Ttr[train])  # Train treatment assignments as a series

        Xte = torch.from_numpy(Xte.astype('float64'))
        Yte = torch.from_numpy(Yte.astype('float64'))
        Tte = torch.from_numpy(Tte.astype('float64'))

        if outcome_model == 'MLP':
            outcome_model = genMLPModule(Xte.shape[1], Xte.shape[1] / 5, 2)
        elif outcome_model == 'linear':
            outcome_model = genLinearModule(Xte.shape[1], 2)
        elif outcome_model == 'CF':
            outcome_model = BalancedNet(Xte.shape[1], Xte.shape[1], 1)

        estimator = HEMM(
            Xte.shape[1],
            comp,
            mu=mu,
            std=std,
            bc=6,
            lamb=0.,
            spread=0.,
            outcome_model=outcome_model,
            epochs=500,
            batch_size=batch_size,
            learning_rate=lr,
            weight_decay=1e-4,
            metric='LL',
            response=response,
            imb_fun='wass'
        )
        estimator.fit(Xtr, Ttr, Ytr, validation_data=(Xdev, Tdev, Ydev))

        Xtr = data['TRAIN']['x'][:, :, i]
        Ttr = data['TRAIN']['t'][:, i]
        Ytr = data['TRAIN']['yf'][:, i]

        Xtr = torch.from_numpy(Xtr.astype('float64'))
        Ytr = torch.from_numpy(Ytr.astype('float64'))
        Ttr = torch.from_numpy(Ttr.astype('float64'))

        in_estimations = estimator.estimate_individual_outcome(Xtr, Ttr)
        out_estimations = estimator.estimate_individual_outcome(Xte, Tte)

        group_proba = estimator.get_groups_proba(Xte)
        self.assertEqual(group_proba.shape, (data['TEST']['x'][:, :, 1].shape[0], comp))

        group_assignment = estimator.get_groups(Xte)
        pd.testing.assert_series_equal(group_assignment, group_proba.idxmax(axis="columns"))

        group_effect = estimator.get_groups_effect(Xte, Tte)
        self.assertEqual(group_effect.shape, (comp,))

        group_sizes = estimator.group_sizes(Xte)
        self.assertEqual(len(group_sizes.keys()), comp)

        return in_estimations, out_estimations

    def test_hemm_estimator(self):
        data = gen_montecarlo(1000, 24, 2)
        in_estimations, out_estimations = self.experiment(data, 1, 3, 'cont', 'CF', 1e-3, 10)
        self.assertEqual(in_estimations.shape, (data['TRAIN']['x'][:, :, 1].shape[0], 2))
        self.assertEqual(out_estimations.shape, (data['TEST']['x'][:, :, 1].shape[0], 2))



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Aug 08, 2018

"""

import unittest

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.linear_model import LinearRegression, LogisticRegression, Ridge
from sklearn.svm import SVC

from causallib.estimation import Standardization, StratifiedStandardization


class TestStandardizationCommon(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        alpha, beta = 5.0, 0.4
        X = pd.Series(np.random.normal(2.0, 1.0, size=100))
        a = pd.Series([0] * 50 + [1] * 50, dtype=np.dtype(int))
        y = X.mul(alpha) + a.mul(beta)
        cls.data_lin = {"X": X.to_frame(), "a": a, "y": y, "alpha": alpha, "beta": beta}
        cls.estimator = None

    def ensure_observed_prediction(self):
        ind_outcome = self.estimator.estimate_individual_outcome(self.data_lin["X"], self.data_lin["a"])
        obs_outcome = np.concatenate((ind_outcome[0][self.data_lin["a"] == 0], ind_outcome[1][self.data_lin["a"] == 1]))
        np.testing.assert_almost_equal(self.data_lin["y"], obs_outcome)

    def ensure_counterfactual_outcomes(self):
        ind_outcome = self.estimator.estimate_individual_outcome(self.data_lin["X"], self.data_lin["a"])
        with self.subTest("Treatment value 0:"):
            np.testing.assert_array_almost_equal(self.data_lin["X"] * self.data_lin["alpha"],
                                                 ind_outcome[0].to_frame())
        with self.subTest("Treatment value 1:"):
            np.testing.assert_array_almost_equal((self.data_lin["X"] * self.data_lin["alpha"]) + self.data_lin["beta"],
                                                 ind_outcome[1].to_frame())

    def ensure_effect_estimation(self):
        with self.subTest("Check by individual effect:"):
            ind_outcome = self.estimator.estimate_individual_outcome(self.data_lin["X"], self.data_lin["a"])
            effect_est = self.estimator.estimate_effect(ind_outcome[1], ind_outcome[0], agg="individual")["diff"]
            np.testing.assert_array_almost_equal(effect_est, np.full_like(effect_est, self.data_lin["beta"]))

        with self.subTest("Check by population effect:"):
            pop_outcome = self.estimator.estimate_population_outcome(self.data_lin["X"], self.data_lin["a"],
                                                                     agg_func="mean")
            effect_est = self.estimator.estimate_effect(pop_outcome[1], pop_outcome[0], agg="population")["diff"]
            self.assertAlmostEqual(effect_est, np.full_like(effect_est, self.data_lin["beta"]), places=5)

    def ensure_pipeline_learner(self):
        from sklearn.preprocessing import StandardScaler, MinMaxScaler
        from sklearn.pipeline import make_pipeline
        learner = make_pipeline(StandardScaler(), MinMaxScaler(), LinearRegression())
        with self.subTest("Test initialization with pipeline learner"):
            self.estimator = self.estimator.__class__(learner)
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test fit with pipeline learner"):
            self.estimator.fit(self.data_lin["X"], self.data_lin["a"], self.data_lin["y"])
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test 'predict' with pipeline learner"):
            self.estimator.estimate_individual_outcome(self.data_lin["X"], self.data_lin["a"])
            self.assertTrue(True)  # Dummy assert for not thrown exception

    def ensure_many_models(self):
        from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor
        from sklearn.neural_network import MLPRegressor
        from sklearn.linear_model import ElasticNet, RANSACRegressor, HuberRegressor, PassiveAggressiveRegressor
        from sklearn.neighbors import KNeighborsRegressor
        from sklearn.svm import SVR, LinearSVR

        import warnings
        from sklearn.exceptions import ConvergenceWarning
        warnings.filterwarnings('ignore', category=ConvergenceWarning)

        for learner in [GradientBoostingRegressor, RandomForestRegressor, MLPRegressor,
                        ElasticNet, RANSACRegressor, HuberRegressor, PassiveAggressiveRegressor,
                        KNeighborsRegressor, SVR, LinearSVR]:
            learner = learner()
            learner_name = str(learner).split("(", maxsplit=1)[0]
            with self.subTest("Test fit using {learner}".format(learner=learner_name)):
                model = self.estimator.__class__(learner)
                model.fit(self.data_lin["X"], self.data_lin["a"], self.data_lin["y"])
                self.assertTrue(True)  # Fit did not crash


class TestStandardization(TestStandardizationCommon):
    @classmethod
    def setUpClass(cls):
        TestStandardizationCommon.setUpClass()
        # Avoids regularization of the model:
        cls.estimator = Standardization(LinearRegression(normalize=True))

    def setUp(self):
        self.estimator.fit(self.data_lin["X"], self.data_lin["a"], self.data_lin["y"])

    def test_is_fitted(self):
        self.assertTrue(hasattr(self.estimator.learner, "coef_"))

    def test_effect_estimation(self):
        with self.subTest("Check by model coefficient:"):
            self.assertAlmostEqual(self.estimator.learner.coef_[0], self.data_lin["beta"], places=5)
        self.ensure_effect_estimation()

    def test_observed_prediction(self):
        self.ensure_observed_prediction()

    def test_counterfactual_outcomes(self):
        self.ensure_counterfactual_outcomes()

    def test_treatment_encoding(self):
        self.estimator = Standardization(LinearRegression(), encode_treatment=True)
        a = self.data_lin["a"].replace({0: "p", 1: "q"})
        self.estimator.fit(self.data_lin["X"], a, self.data_lin["y"])
        with self.subTest("Treatment encoder created:"):
            self.assertTrue(hasattr(self.estimator, "treatment_encoder_"))
        with self.subTest("Treatment categories properly encoded"):
            self.assertSetEqual({"p", "q"}, set(*self.estimator.treatment_encoder_.categories_))
        with self.subTest("Fitted model has the right size"):
            self.assertEqual(len(self.estimator.learner.coef_), self.data_lin["X"].shape[1] + a.nunique())

    def test_pipeline_learner(self):
        self.ensure_pipeline_learner()

    def test_many_models(self):
        self.ensure_many_models()


class TestStandardizationStratified(TestStandardizationCommon):
    @classmethod
    def setUpClass(cls):
        TestStandardizationCommon.setUpClass()
        # Avoids regularization of the model:
        cls.estimator = StratifiedStandardization(LinearRegression(), [0, 1])

    def setUp(self):
        self.estimator.fit(self.data_lin["X"], self.data_lin["a"], self.data_lin["y"])

    def test_is_fitted(self):
        with self.subTest("Test fit for model 0:"):
            self.assertTrue(hasattr(self.estimator.learner[0], "coef_"))
        with self.subTest("Test fit for model 1:"):
            self.assertTrue(hasattr(self.estimator.learner[1], "coef_"))

    def test_effect_estimation(self):
        with self.subTest("Check by model coefficient:"):
            self.assertAlmostEqual(self.estimator.learner[1].intercept_, self.data_lin["beta"], places=5)
            self.assertAlmostEqual(self.estimator.learner[0].intercept_, 0, places=5)
        self.ensure_effect_estimation()

    def test_observed_prediction(self):
        self.ensure_observed_prediction()

    def test_counterfactual_outcomes(self):
        self.ensure_counterfactual_outcomes()

    def test_initialization_without_treatment_values(self):
        estimator = StratifiedStandardization(LinearRegression())
        with self.subTest("Test no-treatment initialization before fit"):
            self.assertIsInstance(estimator.learner, LinearRegression)
        estimator.fit(self.data_lin["X"], self.data_lin["a"], self.data_lin["y"])
        with self.subTest("Test no-treatment initialization after fit"):
            self.assertIsInstance(estimator.learner, dict)
            self.assertSetEqual(set(estimator.learner.keys()), set(self.data_lin["a"]))
        with self.subTest("Test treatment values were added"):
            self.assertIsNone(estimator.treatment_values)  # No treatment values in initiation
            self.assertListEqual([0, 1], estimator.treatment_values_)  # Treatment values added after fit

    def test_initialization_with_dict_of_learners(self):
        learners = {0: LinearRegression(normalize=False),
                    1: Ridge(alpha=5.0)}
        estimator = StratifiedStandardization(learners)
        with self.subTest("Test dictionary-learners initialization before fit"):
            self.assertIsInstance(estimator.learner, dict)
        estimator.fit(self.data_lin["X"], self.data_lin["a"], self.data_lin["y"])
        with self.subTest("Test dictionary-learners keys as expected"):
            self.assertSetEqual(set(estimator.learner.keys()), set(self.data_lin["a"]))
        with self.subTest("Test dictionary-learners types as expected"):
            self.assertIsInstance(estimator.learner, dict)
            self.assertIsInstance(estimator.learner[0], LinearRegression)
            self.assertIsInstance(estimator.learner[1], Ridge)
        with self.subTest("Test dictionary-learners fitted properly"):
            self.assertTrue(hasattr(estimator.learner[0], "coef_"))
            self.assertTrue(hasattr(estimator.learner[1], "coef_"))

    def test_pipeline_learner(self):
        self.ensure_pipeline_learner()

    def test_many_models(self):
        self.ensure_many_models()


class TestStandardizationClassification(TestStandardizationCommon):
    @classmethod
    def setUpClass(cls):
        # Three-class outcome, since decision_function might return a vector when n_classes=2, and we wish to check the
        # matrix form of the output behaves as expected:
        X, y = make_classification(n_features=3, n_informative=2, n_redundant=0, n_repeated=0, n_classes=3,
                                   n_clusters_per_class=1, flip_y=0.0, class_sep=10.0)
        X, a = X[:, :-1], X[:, -1]
        a = (a > np.median(a)).astype(int)
        cls.data_3cls = {"X": pd.DataFrame(X), "a": pd.Series(a), "y": pd.Series(y)}

        # X, y = make_classification(n_features=2, n_informative=1, n_redundant=0, n_repeated=0, n_classes=2,
        #                            n_clusters_per_class=1, flip_y=0.0, class_sep=10.0)
        # X, a = X[:, :-1], X[:, -1]
        # a = (a > np.median(a)).astype(int)
        # cls.data_2cls = {"X": pd.DataFrame(X), "a": pd.Series(a), "y": pd.Series(y)}

    def verify_individual_multiclass_output(self):
        self.estimator.fit(self.data_3cls["X"], self.data_3cls["a"], self.data_3cls["y"])
        ind_outcome = self.estimator.estimate_individual_outcome(self.data_3cls["X"], self.data_3cls["a"])

        with self.subTest("Output size, # samples:"):
            self.assertEqual(self.data_3cls["X"].shape[0], ind_outcome.shape[0])
        with self.subTest("Output size, # predictions:"):
            with self.subTest("Output's multiindex level names are describing treatment and outcome"):
                self.assertEqual(["a", "y"], ind_outcome.columns.names)
            with self.subTest("Output's number of predictions is the same as number of outcome and treatment values"):
                self.assertEqual(self.data_3cls["a"].nunique() * self.data_3cls["y"].nunique(), ind_outcome.shape[1])
                self.assertEqual(self.data_3cls["a"].nunique(), ind_outcome.columns.get_level_values("a").unique().size)
                self.assertEqual(self.data_3cls["y"].nunique(), ind_outcome.columns.get_level_values("y").unique().size)
        return ind_outcome

    def test_predict_proba(self):
        self.estimator = Standardization(LogisticRegression(C=1e6, solver='lbfgs'), predict_proba=True)
        ind_outcome = self.verify_individual_multiclass_output()
        with self.subTest("Test results are probabilities - sum to 1:"):
            for treatment_value, y_pred in ind_outcome.groupby(level="a", axis="columns"):
                pd.testing.assert_series_equal(pd.Series(1.0, index=y_pred.index), y_pred.sum(axis="columns"))

    def test_decision_function(self):
        self.estimator = Standardization(SVC(decision_function_shape='ovr'), predict_proba=True)
        self.verify_individual_multiclass_output()

    def test_predict(self):
        self.estimator = Standardization(LogisticRegression(C=1e6, solver='lbfgs'), predict_proba=False)
        self.estimator.fit(self.data_3cls["X"], self.data_3cls["a"], self.data_3cls["y"])
        ind_outcome = self.estimator.estimate_individual_outcome(self.data_3cls["X"], self.data_3cls["a"])
        with self.subTest("Output size, # predictions:"):
            self.assertEqual(self.data_3cls["a"].nunique(), ind_outcome.shape[1])
            self.assertNotEqual(self.data_3cls["y"].nunique(), ind_outcome.shape[1])



"""
Metrics to assess performance of direct-outcome (counterfactual prediction) models.
Function named as ``*_error`` return a scalar value to minimize:
the lower, the better.

Metrics' interface doesn't strictly follow skleran's metrics interface.
The outcome prediction is expected to be full potential outcome prediction
(one column for each treatment value), and also expects the true treatment assignment.
"""

import numpy as np
import pandas as pd

from causallib.utils.stat_utils import calc_weighted_standardized_mean_differences
from causallib.utils.stat_utils import robust_lookup


def abs_standardized_mean_difference(a, b, **kwargs):
    asmd = calc_weighted_standardized_mean_differences(
        a, b,
        wx=np.ones_like(a),
        wy=np.ones_like(b),
    )
    asmd = np.abs(asmd)
    return asmd


def _get_observed_outcome_prediction(potential_outcomes, a):
    # TODO: duplicated throughout causallib. move to utils or standardization module
    is_predict_proba_classification_result = isinstance(potential_outcomes.columns, pd.MultiIndex)
    if is_predict_proba_classification_result:
        # Classification `outcome_model` with `predict_proba=True` returns
        # a MultiIndex treatment-values (`a`) over outcome-values (`y`)
        # Extract the prediction for the maximal outcome class
        # (probably class `1` in binary classification):
        outcome_values = potential_outcomes.columns.get_level_values(level=-1)
        potential_outcomes = potential_outcomes.xs(
            outcome_values.max(), axis="columns", level=-1, drop_level=True,
        )
    potential_outcomes = robust_lookup(potential_outcomes, a)
    return potential_outcomes


def balanced_residuals_error(
    y_true, y_pred, a_true,
    distance_metric=abs_standardized_mean_difference,
    distance_metric_kwargs=None,
):
    """Computes how different is the residuals distribution of the control group
    from that of the treatment group.
    Residuals are based on the observed (factual) outcome prediction.

    Can plug in any uni-variate two-sample test function.

    Args:
        y_true (pd.Series): The true observed outcomes.
        y_pred (pd.DataFrame): Potential outcome prediction, the output of `estimate_individual_outcome()`.
                               A matrix of (n_samples, n_treatments), with column names as the treatment values.
        a_true (pd.Series): A vector of observed treatment assignment.
        distance_metric (callable): A two sample test function.
            First argument is the residual values of the treatment group, second is for the control group.
            Defaults to absolute standardized mean difference.
        distance_metric_kwargs (dict): Additional keyword arguments needed for the `distance_metric` function.

    Returns:
        score (float):
    """
    if distance_metric_kwargs is None:
        distance_metric_kwargs = {}

    y_pred = _get_observed_outcome_prediction(y_pred, a_true)

    residuals = y_true - y_pred
    treatment_mask = a_true == a_true.max()
    control_mask = a_true == a_true.min()

    score = distance_metric(
        residuals[treatment_mask],
        residuals[control_mask],
        **distance_metric_kwargs,
    )
    return score




# coding: utf-8

# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Sept 25, 2019

import numpy as np
from collections import Counter


def KFoldStratifiedMultiClass(labels, n, seed=0):
    """Returns a stratified n-fold in a dictionary, trying to distribute classes across folds as well as possible."""
    np.random.seed(seed)
    freqs = np.sum(labels, axis=0)  # class frequencies
    tups = zip(freqs, range(19))
    tups = sorted(tups)
    freq_order = map(lambda i: i[1], tups)  # classes sorted by frequency
    folds = {}

    # folds is a dictionary indexed by integer, which contains
    # the indexes of the samples in each fold
    for i in range(n):
        folds[i] = []

    # visited is a set to keep the samples already sorted into folds
    visited = set()
    for j in freq_order:
        occs = np.where(labels[:, j] == 1)[0]  # find the occurrences of the class
        for i, k in enumerate(occs):
            # for each occurrence check if it's been sorted
            if k not in visited:
                # and in case not append it to a given fold
                folds[i % n].append(k)
        # after sorting each class update visited set
        for i in folds.keys():
            visited = visited.union(set(folds[i]))

    # sort remaining examples into folds, trying to keep all folds
    # equally populated.
    for i in range(labels.shape[0]):
        if i not in visited:
            f = np.argmin(map(lambda i: len(folds[i]), range(n)))
            folds[f].append(i)

    return folds


def getMeanandStd(X):
    """Takes the features and computes the mean and std dev of each column."""
    mu = np.mean(X, axis=0)
    std = np.std(X, axis=0)

    return np.array([mu]).astype('float64'), np.array([std]).astype('float64')


def genSplits(T, Y, k=5):
    """Returns k folds indices of the input data."""
    star = ((T * 10) + (Y ** 1)).tolist()
    star = np.vstack((T, Y)).T
    starsum = star.sum(axis=1)
    starsum = ((T * 10) + (Y ** 1))

    Counter(starsum)

    splits = KFoldStratifiedMultiClass(star, k)
    # for split in splits:
    # print(split,len(splits[split]), T[splits[split]].sum(), Y[splits[split]].sum(), Counter(starsum[splits[split]]))

    overlap = set(range(len(T)))
    for split in splits:
        overlap &= set(splits[split])
    # print("Total Overlap:", len(overlap))

    return splits


def returnIndices(splits):
    """Takes the splits and return the train and dev splits."""
    test = []
    train = []
    dev = []

    for split in splits.keys():
        if split in [0]:
            dev += splits[split]
        else:
            train += splits[split]

    return train, dev


def returnIndicesTest(splits):
    """Takes the splits and return the train, test and dev splits."""
    test = []
    train = []
    dev = []

    for split in splits.keys():
        if split in [0, 1]:
            test += splits[split]
        if split in [2]:
            dev += splits[split]
        else:
            train += splits[split]

    return train, test, dev



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jul 22, 2018

"""

import unittest

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression

from causallib.estimation import IPW


class TestIPW(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Data:
        X, a = make_classification(n_features=1, n_informative=1, n_redundant=0, n_repeated=0, n_classes=2,
                                   n_clusters_per_class=1, flip_y=0.0, class_sep=10.0)
        cls.data_r_100 = {"X": pd.DataFrame(X), "a": pd.Series(a)}
        X, a = make_classification(n_features=1, n_informative=1, n_redundant=0, n_repeated=0, n_classes=2,
                                   n_clusters_per_class=1, flip_y=0.2, class_sep=10.0)
        cls.data_r_80 = {"X": pd.DataFrame(X), "a": pd.Series(a)}

        # Data that maps x=0->a=0 and x=1->a=1:
        X = pd.Series([0] * 50 + [1] * 50)
        cls.data_cat_r_100 = {"X": X.to_frame(), "a": X}

        # Data that maps x=0->a=0 and x=1->a=1, but 10% of x=0->a=1 and 10% of x=1->a=0:
        X = pd.Series([0] * 40 + [1] * 10 + [1] * 40 + [0] * 10).to_frame()
        a = pd.Series([0] * 50 + [1] * 50)
        cls.data_cat_r_80 = {"X": X, "a": a}

        # Avoids regularization of the model:
        cls.estimator = IPW(LogisticRegression(C=1e6, solver='lbfgs'), clip_min=0.05, clip_max=0.95,
                            use_stabilized=False)

    def setUp(self):
        self.estimator.fit(self.data_r_100["X"], self.data_r_100["a"])

    def test_is_fitted(self):
        self.assertTrue(hasattr(self.estimator.learner, "coef_"))

    def test_weight_matrix_vector_matching(self):
        a = self.data_r_100["a"]
        p_vec = self.estimator.compute_weights(self.data_r_100["X"], a)
        p_mat = self.estimator.compute_weight_matrix(self.data_r_100["X"], a)
        self.assertEqual(p_vec.size, p_mat.shape[0])
        for i in range(a.shape[0]):
            self.assertAlmostEqual(p_mat.loc[i, a[i]], p_vec[i])

    def test_weight_sizes(self):
        a = self.data_r_100["a"]
        with self.subTest("Weight vector size"):
            p = self.estimator.compute_weights(self.data_r_100["X"], a)
            self.assertEqual(len(p.shape), 1)  # vector has no second axis
            self.assertEqual(p.shape[0], a.shape[0])

        with self.subTest("Weight matrix size"):
            p = self.estimator.compute_weight_matrix(self.data_r_100["X"], a)
            self.assertEqual(len(p.shape), 2)  # Matrix has two dimensions
            self.assertEqual(p.shape[0], a.shape[0])
            self.assertEqual(p.shape[1], np.unique(a).size)

    def ensure_truncation(self, test_weights):
        with self.subTest("Estimator initialization parameters"):
            p = self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"])
            if test_weights:
                p = self.estimator.compute_weights(self.data_r_80["X"], self.data_r_80["a"]).pow(-1)

            self.assertAlmostEqual(p.min(), 0.05)
            self.assertAlmostEqual(p.max(), 1 - 0.05)

        with self.subTest("Overwrite parameters in compute_weights"):
            p = self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.1, clip_max = 0.9)
            if test_weights:
                p = self.estimator.compute_weights(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.1, clip_max=0.9).pow(-1)
            self.assertAlmostEqual(p.min(), 0.1)
            self.assertAlmostEqual(p.max(), 1 - 0.1)

        with self.subTest("Test asymmetric clipping"):
            p = self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.2,
                                                  clip_max=0.9)
            if test_weights:
                p = self.estimator.compute_weights(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.2,
                                                   clip_max=0.9).pow(-1)
            self.assertAlmostEqual(p.min(), 0.2)
            self.assertAlmostEqual(p.max(), 0.9)

        with self.subTest("Test calculation of fraction of clipped observations"):
            probabilities = pd.DataFrame()
            probabilities['col1'] = [0.01, 0.02, 0.03, 0.05, 0.3, 0.6, 0.9, 0.95, 0.99, 0.99]
            probabilities['col2'] = [0.99, 0.98, 0.97, 0.95, 0.7, 0.4, 0.1, 0.05, 0.01, 0.01]
            frac = self.estimator._IPW__count_truncated(probabilities, clip_min=0.05, clip_max=0.95)
            self.assertAlmostEqual(frac, 0.5)

        with self.subTest("Test calculation of fraction of clipped observations - no clipping"):
            probabilities = pd.DataFrame()
            probabilities['col1'] = [0.0, 0.0, 0.0, 1.0, 1.0]
            probabilities['col2'] = [1.0, 1.0, 1.0, 0.0, 0.0]
            frac = self.estimator._IPW__count_truncated(probabilities, clip_min=0.0, clip_max=1.0)
            self.assertAlmostEqual(frac, 0.0)

    def test_weight_truncation(self):
        self.ensure_truncation(test_weights=True)

    def test_propensity_truncation(self):
        self.ensure_truncation(test_weights=False)

        with self.subTest("Illegal truncation values assertion on compute"):
            with self.assertRaises(AssertionError):
                self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.6)
            with self.assertRaises(AssertionError):
                self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"], clip_max=0.4)
            with self.assertRaises(AssertionError):
                self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.6,
                                                  clip_max=0.9)
            with self.assertRaises(AssertionError):
                self.estimator.compute_propensity(self.data_r_80["X"], self.data_r_80["a"], clip_min=0.1,
                                                  clip_max=0.4)

        with self.subTest("Illegal truncation values assertion on initialization"):
            with self.assertRaises(AssertionError):
                IPW(LogisticRegression(), clip_min=0.6)
            with self.assertRaises(AssertionError):
                IPW(LogisticRegression(), clip_max=0.4)
            with self.assertRaises(AssertionError):
                IPW(LogisticRegression(), clip_min=0.1, clip_max=0.4)
            with self.assertRaises(AssertionError):
                IPW(LogisticRegression(), clip_min=0.6, clip_max=0.9)

    def test_weights_sanity_check(self):
        with self.subTest("Linearly separable X should have perfectly predicted propensity score"):
            p = self.estimator.compute_weights(self.data_r_100["X"], self.data_r_100["a"], clip_min=0.0,
                                               clip_max=1.0).pow(-1)
            np.testing.assert_array_almost_equal(p, np.ones_like(p), decimal=3)

        with self.subTest("Train on bijection X|a data and predict on data where q% are flipped"):
            # Train on data that maps x=0->a=0 and x=1->a=1:
            self.estimator.fit(self.data_cat_r_100["X"], self.data_cat_r_100["a"])
            # Predict on a set with mis-mapping: 10% of x=0 have a=1 and 10% of x=1 have a=0:
            p = self.estimator.compute_weights(self.data_cat_r_80["X"], self.data_cat_r_80["a"], clip_min=0.0, clip_max=1.0).pow(
                -1)
            # Extract subjects with mismatching X-a values:
            mis_assigned = np.logical_xor(self.data_cat_r_80["X"].iloc[:, 0], self.data_cat_r_80["a"])
            # See they have the same rate:
            self.assertAlmostEqual(p.mean(), 1.0 - mis_assigned.mean(), 4)
            np.testing.assert_almost_equal(p.mean(), 1.0 - mis_assigned.mean(), decimal=4)

    def test_forcing_probability_learner(self):
        from sklearn.svm import SVC  # Arbitrary model with decision_function instead of predict_proba
        with self.assertRaises(AttributeError):
            IPW(SVC())

    def test_pipeline_learner(self):
        from sklearn.preprocessing import StandardScaler, MinMaxScaler
        from sklearn.pipeline import make_pipeline
        learner = make_pipeline(StandardScaler(), MinMaxScaler(), LogisticRegression(solver='lbfgs'))
        with self.subTest("Test initialization with pipeline learner"):
            self.estimator = IPW(learner)
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test fit with pipeline learner"):
            self.estimator.fit(self.data_r_100["X"], self.data_r_100["a"])
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test 'predict' with pipeline learner"):
            self.estimator.compute_weights(self.data_r_100["X"], self.data_r_100["a"])
            self.assertTrue(True)  # Dummy assert for not thrown exception
